#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################

from pyomo.network import Arc
from idaes.core import (
    FlowsheetBlock,
)

from pyomo.environ import (
    units as pyunits,
    Var,
    assert_optimal_termination,
    value,
)

import idaes.core.util.scaling as iscale
from idaes.core.solvers import get_solver
from idaes.core.util.initialization import propagate_state
from idaes.core.util.model_statistics import degrees_of_freedom
from idaes.models.unit_models import (
    Mixer,
    Separator,
    Product,
    Feed,
)
from idaes.models.unit_models.mixer import MomentumMixingType, MixingType

from pyomo.environ import ConcreteModel, TransformationFactory

from watertap.property_models.multicomp_aq_sol_prop_pack import (
    MCASParameterBlock,
)
from watertap.examples.flowsheets.nf_dspmde import nf

from watertap.costing import WaterTAPCosting

__author__ = "Alexander Dudchenko, Adam Atia"


def main():
    solver = get_solver()
    m = build()
    initialize(m, solver)
    unfix_opt_vars(m)
    nf.add_objective(m)
    results = optimize(m, solver)
    assert_optimal_termination(results)
    print("Optimal cost", value(m.fs.costing.LCOW))
    print("Optimal NF pressure (Bar)", m.fs.NF.pump.outlet.pressure[0].value / 1e5)
    print("Optimal area (m2)", m.fs.NF.nfUnit.area.value)
    print(
        "Optimal nf recovery (%)",
        m.fs.NF.nfUnit.recovery_vol_phase[0.0, "Liq"].value * 100,
    )
    print("bypass (%)", m.fs.by_pass_splitter.split_fraction[0, "bypass"].value * 100)

    print("Feed hardness (mg/L as CaCO3)", m.fs.feed.properties[0].total_hardness.value)
    print(
        "Product hardness (mg/L as CaCO3)",
        m.fs.product.properties[0].total_hardness.value,
    )
    print(
        "Disposal hardness (mg/L as CaCO3)",
        m.fs.disposal.properties[0].total_hardness.value,
    )
    return m


def build():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)
    m.fs.costing = WaterTAPCosting()
    default = nf.define_feed_comp()
    m.fs.properties = MCASParameterBlock(**default)
    m.fs.feed = Feed(property_package=m.fs.properties)
    m.fs.product = Product(property_package=m.fs.properties)
    m.fs.disposal = Product(property_package=m.fs.properties)

    nf.add_hardness_constraint(m.fs.product)

    m.fs.by_pass_splitter = Separator(
        property_package=m.fs.properties,
        outlet_list=["nf_stage", "bypass"],
    )
    # NF UNIT BLOCK
    m.fs.NF = FlowsheetBlock(dynamic=False)

    nf.build_nf_block(m, m.fs.NF)

    m.fs.total_product_mixer = Mixer(
        property_package=m.fs.properties,
        inlet_list=["bypass", "nf_stage"],
        energy_mixing_type=MixingType.none,
        momentum_mixing_type=MomentumMixingType.minimize,
    )
    m.fs.total_product_mixer.mixed_state[0.0].temperature.fix(293.15)
    m.fs.feed_to_splitter = Arc(
        source=m.fs.feed.outlet, destination=m.fs.by_pass_splitter.inlet
    )

    m.fs.splitter_to_nfUnit_feed = Arc(
        source=m.fs.by_pass_splitter.nf_stage, destination=m.fs.NF.feed.inlet
    )

    m.fs.splitter_to_mixer = Arc(
        source=m.fs.by_pass_splitter.bypass, destination=m.fs.total_product_mixer.bypass
    )

    m.fs.nfUnit_product_to_mixer = Arc(
        source=m.fs.NF.product.outlet,
        destination=m.fs.total_product_mixer.nf_stage,
    )

    m.fs.nfUnit_retentate_to_disposal = Arc(
        source=m.fs.NF.retentate.outlet,
        destination=m.fs.disposal.inlet,
    )
    m.fs.mixer_to_product = Arc(
        source=m.fs.total_product_mixer.outlet, destination=m.fs.product.inlet
    )
    m.fs.costing.disposal_cost = Var(
        initialize=0.1,
        bounds=(0, None),
        doc="disposal cost",
        units=pyunits.USD_2020 / pyunits.m**3,
    )
    m.fs.costing.disposal_cost.fix()
    m.fs.costing.register_flow_type("disposal cost", m.fs.costing.disposal_cost)
    m.fs.costing.cost_flow(
        pyunits.convert(
            m.fs.disposal.properties[0].flow_vol_phase["Liq"],
            pyunits.m**3 / pyunits.s,
        ),
        "disposal cost",
    )
    m.fs.costing.cost_process()
    m.fs.costing.add_annual_water_production(m.fs.product.properties[0].flow_vol)
    m.fs.costing.add_LCOW(m.fs.product.properties[0].flow_vol)
    m.fs.costing.add_specific_energy_consumption(m.fs.product.properties[0].flow_vol)
    iscale.set_scaling_factor(m.fs.costing.aggregate_flow_costs["disposal cost"], 1)
    TransformationFactory("network.expand_arcs").apply_to(m)
    return m


def fix_init_vars(m):
    # apply defaults for normal NF init
    nf.fix_init_vars(m)
    # fix initial guess for splitter
    m.fs.by_pass_splitter.split_fraction[0, "bypass"].fix(0.9)
    m.fs.by_pass_splitter.split_fraction[0, "bypass"].setlb(0.05)
    m.fs.by_pass_splitter.split_fraction[0, "bypass"].setub(None)


def initialize(m, solver=None):
    if solver is None:
        solver = get_solver()
    # use standard nf default feed
    nf.set_default_feed(m, solver)
    fix_init_vars(m)

    init_system(m, solver)

    # solve box problem
    print("initialized, DOFs:", degrees_of_freedom(m))
    assert degrees_of_freedom(m) == 0
    results = solver.solve(m, tee=True)
    assert_optimal_termination(results)
    print("Solved box problem")


def init_system(m, solver):
    if solver is None:
        solver = get_solver()
    m.fs.feed.initialize(optarg=solver.options)

    propagate_state(m.fs.feed_to_splitter)

    m.fs.by_pass_splitter.mixed_state.initialize(optarg=solver.options)
    m.fs.by_pass_splitter.initialize(optarg=solver.options)
    propagate_state(m.fs.splitter_to_mixer)
    propagate_state(m.fs.splitter_to_nfUnit_feed)

    nf.init_nf_block(m.fs.NF, solver)

    propagate_state(m.fs.nfUnit_product_to_mixer)

    m.fs.total_product_mixer.mixed_state.initialize(optarg=solver.options)
    m.fs.total_product_mixer.initialize(optarg=solver.options)

    propagate_state(m.fs.nfUnit_retentate_to_disposal)
    propagate_state(m.fs.mixer_to_product)
    m.fs.NF.product.initialize(optarg=solver.options)
    m.fs.NF.retentate.initialize(optarg=solver.options)

    m.fs.costing.initialize()


def unfix_opt_vars(m):
    nf.unfix_opt_vars(m)
    m.fs.by_pass_splitter.split_fraction[0, "bypass"].unfix()


def optimize(m, solver=None):
    if solver is None:
        solver = get_solver()
    result = nf.optimize(m, solver)
    return result


if __name__ == "__main__":
    m = main()



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################

from pyomo.environ import (
    NegativeReals,
    Set,
    Var,
)
from idaes.core import (
    declare_process_block_class,
    FlowDirection,
)
from idaes.core.util import scaling as iscale
from idaes.core.util.exceptions import ConfigurationError
from idaes.core.util.misc import add_object_reference
from idaes.core.base.control_volume0d import ControlVolume0DBlockData
import idaes.logger as idaeslog

from watertap.core.membrane_channel_base import (
    MembraneChannelMixin,
    PressureChangeType,
    CONFIG_Template as Base_CONFIG_Template,
)

CONFIG_Template = Base_CONFIG_Template()


@declare_process_block_class("MembraneChannel0DBlock")
class MembraneChannel0DBlockData(MembraneChannelMixin, ControlVolume0DBlockData):
    def _skip_element(self, x):
        return False

    # overwrite CV0D `add_geometry`
    def add_geometry(
        self, length_var=None, width_var=None, flow_direction=FlowDirection.forward
    ):
        """
        Method to create spatial domain and volume Var in ControlVolume.

        Args:
            length_var - (optional) external variable to use for the length of
                         the channel. If a variable is provided, a
                         reference will be made to this in place of the length
                         Var.

            width_var - (optional) external variable to use for the width of
                         the channel. If a variable is provided, a
                         reference will be made to this in place of the length
                         Var.
            flow_direction - argument indicating direction of material flow
                            relative to length domain. Valid values:
                                - FlowDirection.forward (default), flow goes
                                  from 0 to 1.
                                - FlowDirection.backward, flow goes from 1 to 0
        Returns:
            None
        """
        # Validate and create flow direction attribute, like 1D
        if flow_direction in (flwd for flwd in FlowDirection):
            self._flow_direction = flow_direction
        else:
            raise ConfigurationError(
                "{} invalid value for flow_direction "
                "argument. Must be a FlowDirection Enum.".format(self.name)
            )

        if not hasattr(self, "length") and not hasattr(self, "width"):
            self._add_var_reference(length_var, "length", "length_var")
            self._add_var_reference(width_var, "width", "width_var")

    def add_state_blocks(self, has_phase_equilibrium=None):
        """
        This method constructs the state blocks for the
        control volume.

        Args:
            has_phase_equilibrium: indicates whether equilibrium calculations
                                    will be required in state blocks
        Returns:
            None
        """
        super().add_state_blocks(has_phase_equilibrium=has_phase_equilibrium)

        # quack like a 1D model
        self.length_domain = Set(ordered=True, initialize=(0.0, 1.0))
        add_object_reference(self, "difference_elements", self.length_domain)

        self._set_nfe()

        if self._flow_direction == FlowDirection.forward:
            add_object_reference(
                self,
                "properties",
                {
                    **{
                        (t, 0.0): self.properties_in[t]
                        for t in self.flowsheet().config.time
                    },
                    **{
                        (t, 1.0): self.properties_out[t]
                        for t in self.flowsheet().config.time
                    },
                },
            )
        elif self._flow_direction == FlowDirection.backward:
            add_object_reference(
                self,
                "properties",
                {
                    **{
                        (t, 0): self.properties_out[t]
                        for t in self.flowsheet().config.time
                    },
                    **{
                        (t, 1): self.properties_in[t]
                        for t in self.flowsheet().config.time
                    },
                },
            )
        else:
            raise ConfigurationError(
                "FlowDirection must be set to FlowDirection.forward or FlowDirection.backward."
            )

        self._add_interface_stateblock(has_phase_equilibrium)

    def apply_transformation(self):
        pass

    def _add_pressure_change(self, pressure_change_type=PressureChangeType.calculated):
        if pressure_change_type == PressureChangeType.fixed_per_stage:
            return

        units_meta = self.config.property_package.get_metadata().get_derived_units

        if pressure_change_type == PressureChangeType.fixed_per_unit_length:
            # Pressure change equation when dP/dx = user-specified constant,
            self.dP_dx = Var(
                self.flowsheet().config.time,
                initialize=-5e4,
                bounds=(-2e5, None),
                domain=NegativeReals,
                units=units_meta("pressure") * units_meta("length") ** -1,
                doc="pressure drop per unit length across channel",
            )

            @self.Constraint(
                self.flowsheet().config.time, doc="pressure change due to friction"
            )
            def eq_pressure_change(b, t):
                return b.deltaP[t] == b.dP_dx[t] * b.length

        elif pressure_change_type == PressureChangeType.calculated:
            self.dP_dx = Var(
                self.flowsheet().config.time,
                self.length_domain,
                initialize=-5e4,
                bounds=(-2e5, -1e3),
                domain=NegativeReals,
                units=units_meta("pressure") * units_meta("length") ** -1,
                doc="Pressure drop per unit length of channel at inlet and outlet",
            )

            @self.Constraint(
                self.flowsheet().config.time, doc="Total Pressure drop across channel"
            )
            def eq_pressure_change(b, t):
                return b.deltaP[t] == sum(
                    b.dP_dx[t, x] * b.length / b.nfe for x in b.length_domain
                )

        else:
            raise ConfigurationError(
                f"Unrecognized pressure_change_type {pressure_change_type}"
            )

    def initialize(
        self,
        state_args=None,
        outlvl=idaeslog.NOTSET,
        optarg=None,
        solver=None,
        hold_state=True,
        initialize_guess=None,
    ):
        """
        Initialization routine for the membrane channel control volume

        Keyword Arguments:
            state_args : a dict of arguments to be passed to the property
                         package(s) to provide an initial state for
                         initialization (see documentation of the specific
                         property package) (default = {}).
            outlvl : sets output log level of initialization routine
            optarg : solver options dictionary object (default=None, use
                     default solver options)
            solver : str indicating which solver to use during
                     initialization (default = None)
            hold_state : flag indicating whether the initialization routine
                     should unfix any state variables fixed during
                     initialization, **default** - True. **Valid values:**
                     **True** - states variables are not unfixed, and a dict of
                     returned containing flags for which states were fixed
                     during initialization, **False** - state variables are
                     unfixed after initialization by calling the release_state
                     method.
            initialize_guess : a dict of guesses for solvent_recovery, solute_recovery,
                     and cp_modulus. These guesses offset the initial values
                     for the retentate, permeate, and membrane interface
                     state blocks from the inlet feed
                     (default =
                     {'deltaP': -1e4,
                     'solvent_recovery': 0.5,
                     'solute_recovery': 0.01,
                     'cp_modulus': 1.1})

        Returns:
            If hold_states is True, returns a dict containing flags for which
            states were fixed during initialization.
        """

        # Get inlet state if not provided
        init_log = idaeslog.getInitLogger(self.name, outlvl, tag="control_volume")
        solve_log = idaeslog.getSolveLogger(self.name, outlvl, tag="control_volume")

        # TODO: this function needs to be changed for use on the permeate side
        state_args = self._get_state_args(initialize_guess, state_args)

        # intialize self.properties
        state_args_properties_in = state_args["feed_side"]
        if self._flow_direction == FlowDirection.forward:
            state_args_properties_out = state_args["retentate"]
        else:
            state_args_properties_out = state_args["permeate"]

        source_flags = self.properties_in.initialize(
            state_args=state_args_properties_in,
            outlvl=outlvl,
            optarg=optarg,
            solver=solver,
            hold_state=True,
        )

        self.properties_out.initialize(
            state_args=state_args_properties_out,
            outlvl=outlvl,
            optarg=optarg,
            solver=solver,
        )

        state_args_interface = self._get_state_args_interface(
            initialize_guess, state_args_properties_in, state_args_properties_out
        )
        self.properties_interface.initialize(
            outlvl=outlvl,
            optarg=optarg,
            solver=solver,
            state_args=state_args_interface,
        )

        init_log.info("Initialization Complete")

        if hold_state:
            return source_flags
        else:
            self.release_state(source_flags, outlvl)

    def calculate_scaling_factors(self):
        super().calculate_scaling_factors()

        if hasattr(self, "area"):
            if iscale.get_scaling_factor(self.area) is None:
                iscale.set_scaling_factor(self.area, 100)

        if hasattr(self, "dP_dx"):
            for v in self.dP_dx.values():
                if iscale.get_scaling_factor(v) is None:
                    iscale.set_scaling_factor(v, 1e-4)



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################

import math
import os
import pytest

import pandas as pd

from watertap.examples.flowsheets.lsrro.multi_sweep import run_case

_this_file_path = os.path.dirname(os.path.abspath(__file__))

# NOTE: we used to test up to 5 stages, but those are
#       excluded by the rule below, so no point in running
_test_cases = list(range(1, 3))


@pytest.mark.parametrize("number_of_stages", _test_cases)
@pytest.mark.integration
def test_against_multisweep(number_of_stages, tmp_path):
    csv_file_name = f"{number_of_stages}_stage_results_LSRRO.csv"
    csv_test_file_name = os.path.join("./", csv_file_name)
    csv_baseline_file_name = os.path.join(
        _this_file_path, "parameter_sweep_baselines", csv_file_name
    )
    run_case(number_of_stages, 2, output_filename=csv_test_file_name)

    baseline = pd.read_csv(csv_baseline_file_name).astype(float).T.to_dict()
    test = pd.read_csv(csv_test_file_name).astype(float).T.to_dict()

    assert len(baseline) == len(test)

    for k, base in baseline.items():
        # Don't test those cases which have too many stages
        for s in range(1, number_of_stages + 1):
            if math.isclose(base[f"Membrane Area-Stage {s}"], 1.0, abs_tol=1e-4):
                print(
                    f"Skipping feed concentration {base['# Feed Concentration']}, recovery {base['Volumetric Recovery Rate']}; stage {s} has membrane area of 1"
                )
                break
        else:  # no break
            assert pytest.approx(base, nan_ok=True, rel=1e-02, abs=1e-07) == test[k]



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
from watertap.ui.fsapi import FlowsheetInterface
from watertap.core.util.initialization import assert_degrees_of_freedom
from watertap.examples.flowsheets.case_studies.wastewater_resource_recovery.peracetic_acid_disinfection.peracetic_acid_disinfection import (
    build,
    set_operating_conditions,
    initialize_system,
    solve,
    add_costing,
)
from pyomo.environ import units as pyunits, assert_optimal_termination
from pyomo.util.check_units import assert_units_consistent


def export_to_ui():
    return FlowsheetInterface(
        name="Peracetic acid disinfection",
        do_export=export_variables,
        do_build=build_flowsheet,
        do_solve=solve_flowsheet,
    )


def export_variables(flowsheet=None, exports=None, build_options=None, **kwargs):
    fs = flowsheet
    # --- Input data ---
    # Feed conditions
    exports.add(
        obj=fs.feed.flow_vol[0],
        name="Volumetric flow rate",
        ui_units=pyunits.m**3 / pyunits.hr,
        display_units="m3/h",
        rounding=2,
        description="Inlet volumetric flow rate",
        is_input=True,
        input_category="Feed",
        is_output=True,
        output_category="Feed",
    )
    exports.add(
        obj=fs.feed.conc_mass_comp[0, "peracetic_acid"],
        name="Peracetic acid concentration",
        ui_units=pyunits.mg / pyunits.L,
        display_units="mg/L",
        rounding=2,
        description="Inlet peracetic acid concentration",
        is_input=True,
        input_category="Feed",
        is_output=True,
        output_category="Feed",
    )
    exports.add(
        obj=fs.feed.conc_mass_comp[0, "total_coliforms_fecal_ecoli"],
        name="Ecoli concentration",
        ui_units=pyunits.mg / pyunits.L,
        display_units="mg/L",
        rounding=7,
        description="Inlet total coliforms fecal ecoli concentration",
        is_input=True,
        input_category="Feed",
        is_output=True,
        output_category="Feed",
    )

    # Unit model data, peracetic acid disinfection
    exports.add(
        obj=fs.PAA.recovery_frac_mass_H2O[0],
        name="Water recovery",
        ui_units=pyunits.dimensionless,
        display_units="fraction",  # we should change to %
        rounding=2,
        description="Water recovery [g-H2O treated/g-H2O inlet]",
        is_input=True,
        input_category="Peracetic acid disinfection",
        is_output=False,
    )
    exports.add(
        obj=fs.PAA.reaction_conversion[0, "paa_decomposition"],
        name="PAA decomposition",
        ui_units=pyunits.dimensionless,
        display_units="fraction",  # we should change to %
        rounding=2,
        description="Peracetic acid decomposition [g-PAA decomposed/g-PAA inlet]",
        is_input=True,
        input_category="Peracetic acid disinfection",
        is_output=False,
    )
    exports.add(
        obj=fs.PAA.reaction_conversion[0, "ecoli_inactivation"],
        name="Ecoli inactivation",
        ui_units=pyunits.dimensionless,
        display_units="fraction",  # we should change to %
        rounding=2,
        description="Ecoli (total coliforms fecal ecoli) inactivation [g-ecoli inactivated/g-ecoli inlet]",
        is_input=True,
        input_category="Peracetic acid disinfection",
        is_output=False,
    )
    exports.add(
        obj=fs.PAA.energy_electric_flow_vol_inlet,
        name="Electricity intensity",
        ui_units=pyunits.kWh / pyunits.m**3,
        display_units="fraction",
        rounding=2,
        description="Electricity intensity with respect to inlet flow rate of unit",
        is_input=True,
        input_category="Peracetic acid disinfection",
        is_output=False,
    )
    exports.add(
        obj=fs.PAA.ecoli_cell_mass,
        name="Ecoli mass",
        ui_units=pyunits.kg,
        display_units="kg",
        rounding=16,
        description="Ecoli cell mass",
        is_input=True,
        input_category="Peracetic acid disinfection",
        is_output=False,
    )
    exports.add(
        obj=fs.PAA.disinfection_solution_wt_frac_PAA,
        name="PAA weight fraction of disinfection solution",
        ui_units=pyunits.dimensionless,
        display_units="fraction",
        rounding=3,
        description="Disinfection solution weight fraction of peracetic acid",
        is_input=True,
        input_category="Peracetic acid disinfection",
        is_output=False,
    )
    exports.add(
        obj=fs.PAA.disinfection_solution_density,
        name="Disinfection solution density",
        ui_units=pyunits.kg / pyunits.L,
        display_units="kg/L",
        rounding=3,
        description="Disinfection solution density",
        is_input=True,
        input_category="Peracetic acid disinfection",
        is_output=False,
    )
    exports.add(
        obj=fs.PAA.HRT,
        name="HRT",
        ui_units=pyunits.hr,
        display_units="h",
        rounding=0,
        description="Hydraulic retention time",
        is_input=True,
        input_category="Peracetic acid disinfection",
        is_output=False,
    )

    # Unit cost data, peracetic acid disinfection
    exports.add(
        obj=fs.costing.peracetic_acid_disinfection.sizing_cost[None],
        name="Peracetic acid disinfection cost",
        ui_units=(fs.costing.base_currency * pyunits.day) / pyunits.m**3,
        display_units="($*day)/m3 of PAA",
        rounding=0,
        description="Peracetic acid disinfection capital cost parameter",
        is_input=True,
        input_category="Peracetic acid disinfection costing",
        is_output=False,
    )

    # System costing
    exports.add(
        obj=fs.costing.utilization_factor,
        name="Utilization factor",
        ui_units=pyunits.dimensionless,
        display_units="fraction",
        rounding=2,
        description="Utilization factor - [annual use hours/total hours in year]",
        is_input=True,
        input_category="System costing",
        is_output=False,
    )
    exports.add(
        obj=fs.costing.TIC,
        name="Practical investment factor",
        ui_units=pyunits.dimensionless,
        display_units="fraction",
        rounding=1,
        description="Practical investment factor - [total investment cost/direct "
        "capital costs]",
        is_input=True,
        input_category="System costing",
        is_output=False,
    )
    exports.add(
        obj=fs.costing.plant_lifetime,
        name="Plant lifetime",
        ui_units=pyunits.year,
        display_units="years",
        rounding=1,
        description="Plant lifetime",
        is_input=True,
        input_category="System costing",
        is_output=False,
    )
    exports.add(
        obj=fs.costing.wacc,
        name="Discount rate",
        ui_units=pyunits.dimensionless,
        display_units="fraction",
        rounding=2,
        description="Discount rate used in calculating the capital annualization",
        is_input=True,
        input_category="System costing",
        is_output=False,
    )
    exports.add(
        obj=fs.costing.maintenance_costs_percent_FCI,
        name="Fixed operating cost factor",
        ui_units=1 / pyunits.year,
        display_units="fraction/year",
        rounding=2,
        description="Fixed operating cost factor - [annual fixed operating cost/total "
        "investment cost]",
        is_input=True,
        input_category="System costing",
        is_output=False,
    )
    exports.add(
        obj=fs.costing.electricity_cost,
        name="Electricity cost",
        ui_units=fs.costing.base_currency / pyunits.kWh,
        display_units="$/kWh",
        rounding=3,
        description="Electricity cost",
        is_input=True,
        input_category="System costing",
        is_output=False,
    )
    exports.add(
        obj=fs.costing.disinfection_solution_cost,
        name="Disinfection solution cost",
        ui_units=fs.costing.base_currency / pyunits.gal,
        display_units="$/gal",
        rounding=3,
        description="Disinfection cost",
        is_input=True,
        input_category="System costing",
        is_output=False,
    )

    # Outlets
    exports.add(
        obj=fs.treated_water.properties[0].flow_vol,
        name="Treated water flow rate",
        ui_units=pyunits.m**3 / pyunits.hr,
        display_units="m3/h",
        rounding=2,
        description="Outlet treated water flow rate",
        is_input=False,
        is_output=True,
        output_category="Outlets",
    )
    exports.add(
        obj=fs.treated_water.properties[0].conc_mass_comp["peracetic_acid"],
        name="Treated water PAA concentration",
        ui_units=pyunits.mg / pyunits.L,
        display_units="mg/L",
        rounding=5,
        description="Outlet treated water peracetic acid concentration",
        is_input=False,
        is_output=True,
        output_category="Outlets",
    )
    exports.add(
        obj=fs.treated_water.properties[0].conc_mass_comp[
            "total_coliforms_fecal_ecoli"
        ],
        name="Treated water Ecoli concentration",
        ui_units=pyunits.mg / pyunits.L,
        display_units="mg/L",
        rounding=7,
        description="Outlet treated water ecoli concentration",
        is_input=False,
        is_output=True,
        output_category="Outlets",
    )

    # System metrics
    exports.add(
        obj=fs.costing.LCOT,
        name="LCOT",
        ui_units=fs.costing.base_currency / pyunits.m**3,
        display_units="$/m3 of feed",
        rounding=3,
        description="Levelized cost of treatment including operating and capital costs",
        is_input=False,
        is_output=True,
        output_category="Levelized cost metrics",
    )

    # Normalized metrics
    total_capital_norm = fs.costing.total_capital_cost / fs.feed.properties[0].flow_vol
    exports.add(
        obj=total_capital_norm,
        name="Total capital",
        ui_units=fs.costing.base_currency / (pyunits.m**3 / pyunits.day),
        display_units="$/(m3/day)",
        rounding=1,
        description="Normalized total capital costs accounting for indirect "
        "capital and installation - [total capital costs/feed flow rate]",
        is_input=False,
        is_output=True,
        output_category="Normalized cost metrics",
    )
    direct_capital_norm = (fs.PAA.costing.direct_capital_cost) / fs.feed.properties[
        0
    ].flow_vol
    exports.add(
        obj=direct_capital_norm,
        name="Direct capital",
        ui_units=fs.costing.base_currency / (pyunits.m**3 / pyunits.day),
        display_units="$/(m3/day)",
        rounding=1,
        description="Normalized direct capital costs - [total direct capital "
        "costs/feed flow rate] ",
        is_input=False,
        is_output=True,
        output_category="Normalized cost metrics",
    )
    elec_operating_norm = (
        fs.costing.aggregate_flow_costs["electricity"] / fs.costing.annual_water_inlet
    )
    exports.add(
        obj=elec_operating_norm,
        name="Electricity",
        ui_units=fs.costing.base_currency / pyunits.m**3,
        display_units="$/m3 of feed",
        rounding=2,
        description="Normalized electricity cost - [annual electricity costs/annual "
        "feed flow rate]",
        is_input=False,
        is_output=True,
        output_category="Normalized cost metrics",
    )
    #
    # performance metrics
    recovery_vol = (
        fs.treated_water.properties[0].flow_vol / fs.feed.properties[0].flow_vol
    )
    exports.add(
        obj=recovery_vol,
        name="Volumetric recovery",
        ui_units=pyunits.dimensionless,
        display_units="m3 of product/m3 of feed",
        rounding=3,
        description="Normalized volumetric recovery",
        is_input=False,
        is_output=True,
        output_category="Normalized performance metrics",
    )
    removal_PAA = (
        1
        - fs.treated_water.properties[0].flow_mass_comp["peracetic_acid"]
        / fs.feed.properties[0].flow_mass_comp["peracetic_acid"]
    )
    exports.add(
        obj=removal_PAA,
        name="PAA removal",
        ui_units=pyunits.dimensionless,
        display_units="fraction",
        rounding=3,
        description="Peracetic acid removal fraction [1 - outlet PAA flow/inlet PAA flow]",
        is_input=False,
        is_output=True,
        output_category="Normalized performance metrics",
    )
    removal_Ecoli = (
        1
        - fs.treated_water.properties[0].flow_mass_comp["total_coliforms_fecal_ecoli"]
        / fs.feed.properties[0].flow_mass_comp["total_coliforms_fecal_ecoli"]
    )
    exports.add(
        obj=removal_Ecoli,
        name="Ecoli removal",
        ui_units=pyunits.dimensionless,
        display_units="fraction",
        rounding=3,
        description="Ecoli removal fraction [1 - outlet Ecoli flow/inlet Ecoli flow]",
        is_input=False,
        is_output=True,
        output_category="Normalized performance metrics",
    )

    # Capital costs
    exports.add(
        obj=fs.costing.total_capital_cost,
        name="Total",
        ui_units=fs.costing.base_currency,
        display_units="$",
        rounding=0,
        description="Total capital costs - including investment factor to account "
        "for indirect capital",
        is_input=False,
        is_output=True,
        output_category="Capital costs",
    )
    exports.add(
        obj=fs.PAA.costing.capital_cost,
        name="Peracetic acid disinfection",
        ui_units=fs.costing.base_currency,
        display_units="$",
        rounding=0,
        description="Peracetic acid disinfection",
        is_input=False,
        is_output=True,
        output_category="Capital costs",
    )

    # Operating costs
    exports.add(
        obj=fs.costing.total_operating_cost,
        name="Total",
        ui_units=fs.costing.base_currency / pyunits.year,
        display_units="$/year",
        rounding=0,
        description="Total annual operating costs - including electricity, heating, "
        "and fixed",
        is_input=False,
        is_output=True,
        output_category="Operating costs",
    )
    exports.add(
        obj=fs.costing.aggregate_flow_costs["electricity"],
        name="Electricity",
        ui_units=fs.costing.base_currency / pyunits.year,
        display_units="$/year",
        rounding=0,
        description="Annual electricity costs ",
        is_input=False,
        is_output=True,
        output_category="Operating costs",
    )
    exports.add(
        obj=fs.costing.total_fixed_operating_cost,
        name="Fixed",
        ui_units=fs.costing.base_currency / pyunits.year,
        display_units="$/year",
        rounding=0,
        description="Annual fixed operating costs - these costs include material "
        "replacement, maintenance, and labor",
        is_input=False,
        is_output=True,
        output_category="Operating costs",
    )


def build_flowsheet(build_options=None, **kwargs):
    # build and solve initial flowsheet
    m = build()

    set_operating_conditions(m)
    assert_degrees_of_freedom(m, 0)
    assert_units_consistent(m)

    initialize_system(m)

    results = solve(m)
    assert_optimal_termination(results)

    add_costing(m)
    assert_degrees_of_freedom(m, 0)
    m.fs.costing.initialize()

    results = solve(m)
    assert_optimal_termination(results)
    return m


def solve_flowsheet(flowsheet=None):
    fs = flowsheet
    results = solve(fs)
    return results



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
"""
Tests for ASM2D flowsheet example.

Based on flowsheet from:

Flores-Alsina X., Gernaey K.V. and Jeppsson, U. "Benchmarking biological
nutrient removal in wastewater treatment plants: influence of mathematical model
assumptions", 2012, Wat. Sci. Tech., Vol. 65 No. 8, pp. 1496-1505
"""

# Some more information about this module
__author__ = "Andrew Lee, Alejandro Garciadiego"

import pytest

from pyomo.environ import assert_optimal_termination, value
from pyomo.util.check_units import assert_units_consistent

from idaes.core.util.model_statistics import degrees_of_freedom

from watertap.examples.flowsheets.case_studies.activated_sludge.ASM2D_flowsheet import (
    build_flowsheet,
)


class TestASM2DFlowsheet:
    @pytest.fixture(scope="class")
    def model(self):
        m, res = build_flowsheet()

        m.results = res

        return m

    @pytest.mark.integration
    def test_structure(self, model):
        assert_units_consistent(model)
        assert degrees_of_freedom(model) == 0
        assert_optimal_termination(model.results)

    @pytest.mark.requires_idaes_solver
    @pytest.mark.integration
    def test_results(self, model):
        # Treated water
        assert value(model.fs.Treated.flow_vol[0]) == pytest.approx(0.20921, rel=1e-4)
        assert value(model.fs.Treated.temperature[0]) == pytest.approx(298.15, rel=1e-4)
        assert value(model.fs.Treated.pressure[0]) == pytest.approx(101325, rel=1e-4)
        assert value(model.fs.Treated.conc_mass_comp[0, "S_A"]) == pytest.approx(
            5.4560e-5, rel=1e-4
        )
        assert value(model.fs.Treated.conc_mass_comp[0, "S_F"]) == pytest.approx(
            2.5736e-3, rel=1e-2
        )
        assert value(model.fs.Treated.conc_mass_comp[0, "S_I"]) == pytest.approx(
            30e-3, rel=1e-4
        )
        assert value(model.fs.Treated.conc_mass_comp[0, "S_N2"]) == pytest.approx(
            13.837e-3, abs=1e-4
        )
        assert value(model.fs.Treated.conc_mass_comp[0, "S_NH4"]) == pytest.approx(
            8.98349e-6, rel=1e-4
        )
        assert value(model.fs.Treated.conc_mass_comp[0, "S_NO3"]) == pytest.approx(
            4.3924e-3, abs=1e-4
        )
        assert value(model.fs.Treated.conc_mass_comp[0, "S_O2"]) == pytest.approx(
            7.8744e-3, abs=1e-4
        )
        assert value(model.fs.Treated.conc_mass_comp[0, "S_PO4"]) == pytest.approx(
            2.7958e-3, rel=1e-4
        )
        assert value(model.fs.Treated.conc_mass_comp[0, "X_AUT"]) == pytest.approx(
            1.3277e-3, abs=1e-4
        )
        assert value(model.fs.Treated.conc_mass_comp[0, "X_H"]) == pytest.approx(
            36.331e-3, rel=1e-4
        )
        assert value(model.fs.Treated.conc_mass_comp[0, "X_I"]) == pytest.approx(
            21.1808e-3, rel=1e-4
        )
        assert value(model.fs.Treated.conc_mass_comp[0, "X_MeOH"]) == pytest.approx(
            0, abs=1e-4
        )
        assert value(model.fs.Treated.conc_mass_comp[0, "X_MeP"]) == pytest.approx(
            0, abs=1e-4
        )
        assert value(model.fs.Treated.conc_mass_comp[0, "X_PAO"]) == pytest.approx(
            2.111e-3, abs=1e-4
        )
        assert value(model.fs.Treated.conc_mass_comp[0, "X_PHA"]) == pytest.approx(
            8.981e-5, abs=1e-4
        )
        assert value(model.fs.Treated.conc_mass_comp[0, "X_PP"]) == pytest.approx(
            0.66913e-3, abs=1e-4
        )
        assert value(model.fs.Treated.conc_mass_comp[0, "X_S"]) == pytest.approx(
            1.15897e-3, rel=1e-4
        )
        assert value(model.fs.Treated.conc_mass_comp[0, "X_TSS"]) == pytest.approx(
            76.489e-3, rel=1e-4
        )
        assert value(model.fs.Treated.alkalinity[0]) == pytest.approx(
            7.243e-3, rel=1e-4
        )
        # Sludge stream
        assert value(model.fs.Sludge.flow_vol[0]) == pytest.approx(4.2808e-3, rel=1e-4)
        assert value(model.fs.Sludge.temperature[0]) == pytest.approx(298.15, rel=1e-4)
        assert value(model.fs.Sludge.pressure[0]) == pytest.approx(101325, rel=1e-4)
        assert value(model.fs.Sludge.conc_mass_comp[0, "S_A"]) == pytest.approx(
            5.4560e-5, rel=1e-4
        )
        assert value(model.fs.Sludge.conc_mass_comp[0, "S_F"]) == pytest.approx(
            2.5736e-3, rel=1e-2
        )
        assert value(model.fs.Sludge.conc_mass_comp[0, "S_I"]) == pytest.approx(
            30e-3, rel=1e-4
        )
        assert value(model.fs.Sludge.conc_mass_comp[0, "S_N2"]) == pytest.approx(
            13.837e-3, abs=1e-4
        )
        assert value(model.fs.Sludge.conc_mass_comp[0, "S_NH4"]) == pytest.approx(
            8.9834e-6, rel=1e-4
        )
        assert value(model.fs.Sludge.conc_mass_comp[0, "S_NO3"]) == pytest.approx(
            4.3924e-3, abs=1e-4
        )
        assert value(model.fs.Sludge.conc_mass_comp[0, "S_O2"]) == pytest.approx(
            7.8744e-3, abs=1e-4
        )
        assert value(model.fs.Sludge.conc_mass_comp[0, "S_PO4"]) == pytest.approx(
            2.7958e-3, rel=1e-4
        )
        assert value(model.fs.Sludge.conc_mass_comp[0, "X_AUT"]) == pytest.approx(
            58.6744e-3, abs=1e-4
        )
        assert value(model.fs.Sludge.conc_mass_comp[0, "X_H"]) == pytest.approx(
            1.6200, rel=1e-4
        )
        assert value(model.fs.Sludge.conc_mass_comp[0, "X_I"]) == pytest.approx(
            953.686e-3, rel=1e-4
        )
        assert value(model.fs.Sludge.conc_mass_comp[0, "X_MeOH"]) == pytest.approx(
            0, abs=1e-4
        )
        assert value(model.fs.Sludge.conc_mass_comp[0, "X_MeP"]) == pytest.approx(
            0, abs=1e-4
        )
        assert value(model.fs.Sludge.conc_mass_comp[0, "X_PAO"]) == pytest.approx(
            93.795e-3, abs=1e-4
        )
        assert value(model.fs.Sludge.conc_mass_comp[0, "X_PHA"]) == pytest.approx(
            4.0913e-3, abs=1e-4
        )
        assert value(model.fs.Sludge.conc_mass_comp[0, "X_PP"]) == pytest.approx(
            30.522e-3, abs=1e-4
        )
        assert value(model.fs.Sludge.conc_mass_comp[0, "X_S"]) == pytest.approx(
            50.995e-3, rel=1e-4
        )
        assert value(model.fs.Sludge.conc_mass_comp[0, "X_TSS"]) == pytest.approx(
            3.4078, rel=1e-4
        )
        assert value(model.fs.Sludge.alkalinity[0]) == pytest.approx(
            7.2430e-3, rel=1e-4
        )



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
"""
This module contains a zero-order representation of a backwash solids handling unit.
"""

from idaes.core import declare_process_block_class
from pyomo.environ import Reference
from watertap.core import build_sido, pump_electricity, ZeroOrderBaseData

# Some more information about this module
__author__ = "Adam Atia"


@declare_process_block_class("BackwashSolidsHandlingZO")
class BackwashSolidsHandlingZOData(ZeroOrderBaseData):
    """
    Zero-Order model for a Backwash Solids Handling unit operation.
    """

    CONFIG = ZeroOrderBaseData.CONFIG()

    def build(self):
        super().build()

        self._tech_type = "backwash_solids_handling"

        build_sido(self)
        self._Q = Reference(self.properties_in[:].flow_vol)
        pump_electricity(self, self._Q)



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
"""
Feed block for zero-order flowsheets with methods for getting concentration
data from database.
"""

from pyomo.environ import (
    check_optimal_termination,
    Constraint,
    units as pyunits,
    value,
    Var,
)

from idaes.models.unit_models.feed import FeedData
from idaes.core import declare_process_block_class
import idaes.logger as idaeslog
from idaes.core.solvers import get_solver
from idaes.core.util.exceptions import InitializationError

from watertap.core import InitializationMixin

# Some more inforation about this module
__author__ = "Andrew Lee"

# Set up logger
_log = idaeslog.getLogger(__name__)


@declare_process_block_class("FeedZO")
class FeedZOData(InitializationMixin, FeedData):
    """
    Zero-Order feed block.
    """

    CONFIG = FeedData.CONFIG()

    def build(self):
        super().build()

        units = self.config.property_package.get_metadata().get_derived_units
        comp_list = self.config.property_package.solute_set

        self.flow_vol = Var(
            self.flowsheet().time,
            initialize=1,
            units=units("volume") / units("time"),
            doc="Volumetric flowrate in feed",
        )

        self.conc_mass_comp = Var(
            self.flowsheet().time,
            comp_list,
            initialize=1,
            units=units("density_mass"),
            doc="Component mass concentrations",
        )

        def rule_Q(blk, t):
            return self.flow_vol[t] * self.properties[t].dens_mass == sum(
                self.properties[t].flow_mass_comp[j]
                for j in self.properties[t].component_list
            )

        self.flow_vol_constraint = Constraint(
            self.flowsheet().time, rule=rule_Q, doc="Volumetric flowrate of the feed"
        )

        def rule_C(blk, t, j):
            return (
                self.conc_mass_comp[t, j]
                * sum(
                    self.properties[t].flow_mass_comp[k]
                    for k in self.properties[t].component_list
                )
                == self.properties[t].flow_mass_comp[j] * self.properties[t].dens_mass
            )

        self.conc_mass_constraint = Constraint(
            self.flowsheet().time,
            comp_list,
            rule=rule_C,
            doc="Component mass concentrations",
        )

    def load_feed_data_from_database(self, overwrite=False):
        """
        Method to load initial flowrate and concentrations from database.

        Args:
            overwrite - (default = False), indicates whether fixed values
                        should be overwritten by values from database or not.

        Returns:
            None

        Raises:
            KeyError if flowrate or concentration values not defined in data
        """
        # Get database and water source from property package
        db = self.config.property_package.config.database
        water_source = self.config.property_package.config.water_source

        # Get feed data from database
        data = db.get_source_data(water_source)

        for t in self.flowsheet().time:
            if overwrite or not self.flow_vol[t].fixed:
                try:
                    val = data["default_flow"]["value"]
                    units = getattr(pyunits, data["default_flow"]["units"])
                    self.flow_vol[t].fix(val * units)
                except KeyError:
                    _log.info(
                        f"{self.name} no default flowrate was defined "
                        f"in database water source. Value was not fixed."
                    )

        for (t, j), v in self.conc_mass_comp.items():
            if overwrite or not v.fixed:
                try:
                    val = data["solutes"][j]["value"]
                    units = getattr(pyunits, data["solutes"][j]["units"])
                    v.fix(val * units)
                except KeyError:
                    _log.info(
                        f"{self.name} component {j} was not defined in "
                        f"database water source. Value was not fixed."
                    )

        # Set initial values for mass flows in properties based on these
        # Assuming density of 1000
        for t in self.flowsheet().time:
            for j in self.properties[t].params.solute_set:
                self.properties[t].flow_mass_comp[j].set_value(
                    value(self.flow_vol[t] * self.conc_mass_comp[t, j])
                )
            self.properties[t].flow_mass_comp["H2O"].set_value(
                value(self.flow_vol[t] * 1000)
            )

    def initialize_build(
        self, state_args=None, outlvl=idaeslog.NOTSET, solver=None, optarg=None
    ):
        """
        This method calls the initialization method of the Feed state block.

        Keyword Arguments:
            state_args : a dict of arguments to be passed to the property
                           package(s) to provide an initial state for
                           initialization (see documentation of the specific
                           property package) (default = None).
            outlvl : sets output level of initialization routine
            optarg : solver options dictionary object (default=None, use
                     default solver options)
            solver : str indicating which solver to use during
                     initialization (default = None, use default solver)

        Returns:
            None
        """
        # ---------------------------------------------------------------------
        init_log = idaeslog.getInitLogger(self.name, outlvl, tag="unit")
        solve_log = idaeslog.getSolveLogger(self.name, outlvl, tag="unit")

        if optarg is None:
            optarg = {}

        opt = get_solver(solver, optarg)

        if state_args is None:
            state_args = {}

        # Initialize state block
        self.properties.initialize(
            outlvl=outlvl, optarg=optarg, solver=solver, state_args=state_args
        )

        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            res = opt.solve(self, tee=slc.tee)
        init_log.info("Initialization complete: {}.".format(idaeslog.condition(res)))

        if not check_optimal_termination(res):
            raise InitializationError(
                f"{self.name} failed to initialize successfully. Please check "
                f"the output logs for more information."
            )

    # no costing method
    @property
    def default_costing_method(self):
        return lambda *args, **kwargs: None



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
from watertap.tools.parameter_sweep import (
    LinearSample,
    parameter_sweep,
)
import watertap.examples.flowsheets.case_studies.wastewater_resource_recovery.supercritical_sludge_to_gas.supercritical_sludge_to_gas as supercritical_sludge_to_gas


def set_up_sensitivity(m):
    outputs = {}
    optimize_kwargs = {"fail_flag": False}
    opt_function = supercritical_sludge_to_gas.solve

    # create outputs
    outputs["LCOW"] = m.fs.costing.LCOW
    outputs["LCOG"] = m.fs.costing.LCOG
    outputs["LCOC"] = m.fs.costing.LCOC
    outputs["LCOS"] = m.fs.costing.LCOS

    return outputs, optimize_kwargs, opt_function


def run_analysis(case_num=1, nx=11, interpolate_nan_outputs=True, save_outputs=None):

    m = supercritical_sludge_to_gas.main()[0]

    outputs, optimize_kwargs, opt_function = set_up_sensitivity(m)

    sweep_params = {}
    if case_num == 1:
        m.fs.costing.electricity_cost.unfix()
        sweep_params["electricity_cost"] = LinearSample(
            m.fs.costing.electricity_cost, 0.07, 0.15, nx
        )
    elif case_num == 2:
        m.fs.costing.catalyst_ATHTL_cost.unfix()
        sweep_params["catalyst_ATHTL_cost"] = LinearSample(
            m.fs.costing.catalyst_ATHTL_cost, 0.3, 0.5, nx
        )
    elif case_num == 3:
        m.fs.costing.catalyst_HTG_cost.unfix()
        sweep_params["catalyst_HTG_cost"] = LinearSample(
            m.fs.costing.catalyst_HTG_cost, 0.5, 30, nx
        )
    else:
        raise ValueError("case_num = %d not recognized." % (case_num))

    global_results = parameter_sweep(
        m,
        sweep_params,
        outputs,
        csv_results_file_name=save_outputs,
        optimize_function=opt_function,
        optimize_kwargs=optimize_kwargs,
        interpolate_nan_outputs=interpolate_nan_outputs,
    )

    return global_results, sweep_params, m


if __name__ == "__main__":
    results, sweep_params, m = run_analysis()



###############################################################################
# WaterTAP Copyright (c) 2021, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National
# Laboratory, National Renewable Energy Laboratory, and National Energy
# Technology Laboratory (subject to receipt of any required approvals from
# the U.S. Dept. of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#
###############################################################################
"""
CSTR unit model for BSM2 and plant-wide wastewater treatment modeling.
This unit inherits from the IDAES CSTR unit.
"""

# Import IDAES cores
from idaes.core import (
    declare_process_block_class,
)
from idaes.models.unit_models.cstr import CSTRData as CSTRIDAESData

import idaes.logger as idaeslog

from pyomo.environ import (
    Constraint,
    NonNegativeReals,
    Var,
    units as pyunits,
)

from watertap.costing.unit_models.cstr import cost_cstr

__author__ = "Marcus Holly"


# Set up logger
_log = idaeslog.getLogger(__name__)


@declare_process_block_class("CSTR")
class CSTRData(CSTRIDAESData):
    """
    CSTR unit block for BSM2
    """

    CONFIG = CSTRIDAESData.CONFIG()

    def build(self):
        """
        Begin building model.
        Args:
            None
        Returns:
            None
        """

        # Call UnitModel.build to set up dynamics
        super(CSTRData, self).build()

        self.hydraulic_retention_time = Var(
            self.flowsheet().time,
            initialize=4,
            domain=NonNegativeReals,
            units=pyunits.s,
            doc="Hydraulic retention time",
        )

        def CSTR_retention_time_rule(self, t):
            return (
                self.hydraulic_retention_time[t]
                == self.volume[t] / self.control_volume.properties_in[t].flow_vol
            )

        self.CSTR_retention_time = Constraint(
            self.flowsheet().time,
            rule=CSTR_retention_time_rule,
            doc="Total CSTR retention time",
        )

    @property
    def default_costing_method(self):
        return cost_cstr



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################

import logging

import pyomo.environ as pyo
from pyomo.core.base.block import _BlockData
from pyomo.core.kernel.block import IBlock
from pyomo.solvers.plugins.solvers.IPOPT import IPOPT

import idaes.core.util.scaling as iscale
from idaes.core.util.scaling import (
    get_scaling_factor,
    set_scaling_factor,
    unset_scaling_factor,
)
from idaes.logger import getLogger

_log = getLogger("watertap.core")

_pyomo_nl_writer_log = logging.getLogger("pyomo.repn.plugins.nl_writer")


def _pyomo_nl_writer_logger_filter(record):
    msg = record.getMessage()
    if "scaling_factor" in msg and "model contains export suffix" in msg:
        return False
    return True


@pyo.SolverFactory.register(
    "ipopt-watertap",
    doc="The Ipopt NLP solver, with user-based variable and automatic Jacobian constraint scaling",
)
class IpoptWaterTAP(IPOPT):
    def __init__(self, **kwds):
        kwds["name"] = "ipopt-watertap"
        self._cleanup_needed = False
        super().__init__(**kwds)

    def _presolve(self, *args, **kwds):
        if len(args) > 1 or len(args) == 0:
            raise TypeError(
                f"IpoptWaterTAP.solve takes 1 positional argument but {len(args)} were given"
            )
        if not isinstance(args[0], (_BlockData, IBlock)):
            raise TypeError(
                "IpoptWaterTAP.solve takes 1 positional argument: a Pyomo ConcreteModel or Block"
            )

        # until proven otherwise
        self._cleanup_needed = False

        self._tee = kwds.get("tee", False)

        # Set the default watertap options
        if "tol" not in self.options:
            self.options["tol"] = 1e-08
        if "constr_viol_tol" not in self.options:
            self.options["constr_viol_tol"] = 1e-08
        if "bound_relax_factor" not in self.options:
            self.options["bound_relax_factor"] = 0.0
        if "honor_original_bounds" not in self.options:
            self.options["honor_original_bounds"] = "no"

        if not self._is_user_scaling():
            super()._presolve(*args, **kwds)
            self._cleanup()
            return

        if self._tee:
            print(
                "ipopt-watertap: Ipopt with user variable scaling and IDAES jacobian constraint scaling"
            )

        # These options are typically available with gradient-scaling, and they
        # have corresponding options in the IDAES constraint_autoscale_large_jac
        # function. Here we use their Ipopt names and default values, see
        # https://coin-or.github.io/Ipopt/OPTIONS.html#OPT_NLP_Scaling
        max_grad = self._get_option("nlp_scaling_max_gradient", 100)
        min_scale = self._get_option("nlp_scaling_min_value", 1e-08)

        # These options are custom for the IDAES constraint_autoscale_large_jac
        # function. We expose them as solver options as this has become part
        # of the solve process.
        ignore_variable_scaling = self._get_option("ignore_variable_scaling", False)
        ignore_constraint_scaling = self._get_option("ignore_constraint_scaling", False)

        self._model = args[0]
        self._cache_scaling_factors()
        self._cleanup_needed = True
        _pyomo_nl_writer_log.addFilter(_pyomo_nl_writer_logger_filter)

        # NOTE: This function sets the scaling factors on the
        #       constraints. Hence we cache the constraint scaling
        #       factors and reset them to their original values
        #       so that repeated calls to solve change the scaling
        #       each time based on the initial values, just like in Ipopt.
        try:
            _, _, nlp = iscale.constraint_autoscale_large_jac(
                self._model,
                ignore_constraint_scaling=ignore_constraint_scaling,
                ignore_variable_scaling=ignore_variable_scaling,
                max_grad=max_grad,
                min_scale=min_scale,
            )
        except Exception as err:
            nlp = None
            if str(err) == "Error in AMPL evaluation":
                print(
                    "ipopt-watertap: Issue in AMPL function evaluation; Jacobian constraint scaling not applied."
                )
                halt_on_ampl_error = self.options.get("halt_on_ampl_error", "yes")
                if halt_on_ampl_error == "no":
                    print(
                        "ipopt-watertap: halt_on_ampl_error=no, so continuing with optimization."
                    )
                else:
                    self._cleanup()
                    raise RuntimeError(
                        "Error in AMPL evaluation.\n"
                        "Run ipopt with halt_on_ampl_error=yes and symbolic_solver_labels=True to see the affected function."
                    )
            else:
                print("Error in constraint_autoscale_large_jac")
                self._cleanup()
                raise

        # set different default for `alpha_for_y` if this is an LP
        # see: https://coin-or.github.io/Ipopt/OPTIONS.html#OPT_alpha_for_y
        if nlp is not None:
            if nlp.nnz_hessian_lag() == 0:
                if "alpha_for_y" not in self.options:
                    self.options["alpha_for_y"] = "bound-mult"

        try:
            # this creates the NL file, among other things
            return super()._presolve(*args, **kwds)
        except:
            self._cleanup()
            raise

    def _cleanup(self):
        if self._cleanup_needed:
            self._reset_scaling_factors()
            # remove our reference to the model
            del self._model
            _pyomo_nl_writer_log.removeFilter(_pyomo_nl_writer_logger_filter)

    def _postsolve(self):
        self._cleanup()
        return super()._postsolve()

    def _cache_scaling_factors(self):
        self._scaling_cache = [
            (c, get_scaling_factor(c))
            for c in self._model.component_data_objects(
                pyo.Constraint, active=True, descend_into=True
            )
        ]

    def _reset_scaling_factors(self):
        for c, s in self._scaling_cache:
            if s is None:
                unset_scaling_factor(c)
            else:
                set_scaling_factor(c, s)
        del self._scaling_cache

    def _get_option(self, option_name, default_value):
        # NOTE: options get reset to their original value at the end of the
        #       OptSolver.solve. The options in _presolve (where this is called)
        #       are already copies of the original, so it is safe to pop them so
        #       they don't get sent to Ipopt.
        option_value = self.options.pop(option_name, None)
        if option_value is None:
            option_value = default_value
        else:
            if self._tee:
                print(f"ipopt-watertap: {option_name}={option_value}")
        return option_value

    def _is_user_scaling(self):
        if "nlp_scaling_method" not in self.options:
            self.options["nlp_scaling_method"] = "user-scaling"
        if self.options["nlp_scaling_method"] != "user-scaling":
            if self._tee:
                print(
                    "The ipopt-watertap solver is designed to be run with user-scaling. "
                    f"Ipopt with nlp_scaling_method={self.options['nlp_scaling_method']} will be used instead"
                )
            return False
        return True


## reconfigure IDAES to use the ipopt-watertap solver
import idaes

_default_solver_config_value = idaes.cfg.get("default_solver")
_idaes_default_solver = _default_solver_config_value._default

_default_solver_config_value.set_default_value("ipopt-watertap")
if not _default_solver_config_value._userSet:
    _default_solver_config_value.reset()



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################

# Import Pyomo libraries
from pyomo.environ import (
    Var,
    check_optimal_termination,
    Param,
    Suffix,
    NonNegativeReals,
    value,
    exp,
    log10,
    units as pyunits,
)

from pyomo.common.config import Bool, ConfigBlock, ConfigValue, In

# Import IDAES cores
from idaes.core import (
    declare_process_block_class,
    EnergyBalanceType,
    MaterialBalanceType,
    MomentumBalanceType,
    UnitModelBlockData,
    useDefault,
)
from idaes.core.util.constants import Constants
from idaes.core.solvers import get_solver
from idaes.core.util.config import is_physical_parameter_block
from idaes.core.util.exceptions import ConfigurationError, InitializationError
import idaes.core.util.scaling as iscale
import idaes.logger as idaeslog

from watertap.core import ControlVolume0DBlock, InitializationMixin

__author__ = "Austin Ladshaw"

_log = idaeslog.getLogger(__name__)

# Name of the unit model
@declare_process_block_class("BoronRemoval")
class BoronRemovalData(InitializationMixin, UnitModelBlockData):
    """
    0D Boron Removal model for after 1st Stage of RO

    This model is an approximate equilibrium reactor wherein it is
    assumed that...

    (1) All reactions and activities are assumed ideal

    (2) Only major reactions are water dissociation
    and boron dissociation

    (3) Only 1 caustic chemical is being added to raise pH

    (4) The caustic additive will always completely dissociate into
    a cation and some amount hydroxide anions
    (e.g., NaOH --> Na+ + OH-, Ca(OH2) --> Ca2+ + 2 OH-, etc)

    (5) Any other ions remaining in solution do not significantly
    change with pH changes, but do help act as buffers to
    changes in pH (i.e., will absorb/contribute protons
    proportional to their charge).
    """

    # CONFIG are options for the unit model
    CONFIG = ConfigBlock()

    CONFIG.declare(
        "dynamic",
        ConfigValue(
            domain=In([False]),
            default=False,
            description="Dynamic model flag; must be False",
            doc="""Indicates whether this model will be dynamic or not,
    **default**: False. The filtration unit does not support dynamic
    behavior, thus this must be False.""",
        ),
    )

    CONFIG.declare(
        "has_holdup",
        ConfigValue(
            default=False,
            domain=In([False]),
            description="Holdup construction flag; must be False",
            doc="""Indicates whether holdup terms should be constructed or not.
    **default**: False. The filtration unit does not have defined volume, thus
    this must be False.""",
        ),
    )

    CONFIG.declare(
        "material_balance_type",
        ConfigValue(
            default=MaterialBalanceType.useDefault,
            domain=In(MaterialBalanceType),
            description="Material balance construction flag",
            doc="""Indicates what type of mass balance should be constructed,
    **default** - MaterialBalanceType.useDefault.
    **Valid values:** {
    **MaterialBalanceType.useDefault - refer to property package for default
    balance type
    **MaterialBalanceType.none** - exclude material balances,
    **MaterialBalanceType.componentPhase** - use phase component balances,
    **MaterialBalanceType.componentTotal** - use total component balances,
    **MaterialBalanceType.elementTotal** - use total element balances,
    **MaterialBalanceType.total** - use total material balance.}""",
        ),
    )

    CONFIG.declare(
        "is_isothermal",
        ConfigValue(
            default=True,
            domain=Bool,
            description="""Assume isothermal conditions for control volume(s); energy_balance_type must be EnergyBalanceType.none,
    **default** - True.""",
        ),
    )

    CONFIG.declare(
        "energy_balance_type",
        ConfigValue(
            default=EnergyBalanceType.none,
            domain=In(EnergyBalanceType),
            description="Energy balance construction flag",
            doc="""Indicates what type of energy balance should be constructed,
    **default** - EnergyBalanceType.none.
    **Valid values:** {
    **EnergyBalanceType.useDefault - refer to property package for default
    balance type
    **EnergyBalanceType.none** - exclude energy balances,
    **EnergyBalanceType.enthalpyTotal** - single enthalpy balance for material,
    **EnergyBalanceType.enthalpyPhase** - enthalpy balances for each phase,
    **EnergyBalanceType.energyTotal** - single energy balance for material,
    **EnergyBalanceType.energyPhase** - energy balances for each phase.}""",
        ),
    )

    CONFIG.declare(
        "momentum_balance_type",
        ConfigValue(
            default=MomentumBalanceType.pressureTotal,
            domain=In(MomentumBalanceType),
            description="Momentum balance construction flag",
            doc="""Indicates what type of momentum balance should be constructed,
    **default** - MomentumBalanceType.pressureTotal.
    **Valid values:** {
    **MomentumBalanceType.none** - exclude momentum balances,
    **MomentumBalanceType.pressureTotal** - single pressure balance for material,
    **MomentumBalanceType.pressurePhase** - pressure balances for each phase,
    **MomentumBalanceType.momentumTotal** - single momentum balance for material,
    **MomentumBalanceType.momentumPhase** - momentum balances for each phase.}""",
        ),
    )

    CONFIG.declare(
        "property_package",
        ConfigValue(
            default=useDefault,
            domain=is_physical_parameter_block,
            description="Property package to use for control volume",
            doc="""Property parameter object used to define property calculations,
    **default** - useDefault.
    **Valid values:** {
    **useDefault** - use default package from parent model or flowsheet,
    **PhysicalParameterObject** - a PhysicalParameterBlock object.}""",
        ),
    )

    CONFIG.declare(
        "property_package_args",
        ConfigBlock(
            implicit=True,
            description="Arguments to use for constructing property packages",
            doc="""A ConfigBlock with arguments to be passed to a property block(s)
    and used when constructing these,
    **default** - None.
    **Valid values:** {
    see property package for documentation.}""",
        ),
    )

    CONFIG.declare(
        "chemical_mapping_data",
        ConfigValue(
            default={},
            domain=dict,
            description="Dictionary of chemical species names and their mapping",
            doc="""
        Dictionary of chemical species names from the property
        package and how they map to specific species needed for solving a simple
        boron speciation problem in an equilibrium reactor. This dictionary must
        have the following format [Required]::

            {'boron_name': 'name_of_species_representing_boron', #[is required]
            'borate_name': 'name_of_species_representing_borate', #[is required]
            'proton_name': 'name_of_species_representing_protons',  #[is optional]
            'hydroxide_name': 'name_of_species_representing_hydroxides', #[is optional]
            'caustic_additive':
                {
                    'additive_name': 'name_of_the_actual_chemical', #[is optional]
                    'cation_name': 'name_of_cation_species_in_additive', #[is required]
                    'mw_additive': (value, units), #[is required]
                    'moles_cation_per_additive': value, #[is required]
                },
            }

            """,
        ),
    )

    def _validate_config(self):
        common_msg = (
            "The 'chemical_mapping_data' dict MUST contain a dict of names that map \n"
            + "to each chemical name in the property package for boron and borate. \n"
            + "Optionally, user may provide names that also map to protons and hydroxide, \n"
            + "as well as to the cation from the caustic additive. The 'caustic_additive' \n"
            + "must be a dict that contains molecular weight and charge.\n\n"
            + "Example:\n"
            + "-------\n"
            + "{'boron_name': 'B[OH]3',    #[is required]\n"
            + " 'borate_name': 'B[OH]4_-', #[is required]\n"
            + " 'proton_name': 'H_+',      #[is OPTIONAL]\n"
            + " 'hydroxide_name': 'OH_-',  #[is OPTIONAL]\n"
            + " 'caustic_additive': \n"
            + "     {'additive_name': 'NaOH',                    #[is OPTIONAL]\n"
            + "      'cation_name': 'Na_+',                      #[is required]\n"
            + "      'mw_additive': (40, pyunits.g/pyunits.mol), #[is required]\n"
            + "      'moles_cation_per_additive': 1,             #[is required]\n"
            + "     }, \n"
            + "}\n\n"
        )
        if (
            type(self.config.chemical_mapping_data) != dict
            or self.config.chemical_mapping_data == {}
        ):
            raise ConfigurationError(
                "\n\n Did not provide a 'dict' for 'chemical_mapping_data' \n"
                + common_msg
            )
        if (
            "boron_name" not in self.config.chemical_mapping_data
            or "borate_name" not in self.config.chemical_mapping_data
            or "caustic_additive" not in self.config.chemical_mapping_data
        ):
            raise ConfigurationError(
                "\n\n Missing some required information in 'chemical_mapping_data' \n"
                + common_msg
            )
        if (
            "mw_additive" not in self.config.chemical_mapping_data["caustic_additive"]
            or "moles_cation_per_additive"
            not in self.config.chemical_mapping_data["caustic_additive"]
            or "cation_name"
            not in self.config.chemical_mapping_data["caustic_additive"]
        ):
            raise ConfigurationError(
                "\n\n Missing some required information in 'chemical_mapping_data' \n"
                + common_msg
            )
        if (
            type(self.config.chemical_mapping_data["caustic_additive"]["mw_additive"])
            != tuple
        ):
            raise ConfigurationError(
                "\n Did not provide a tuple for 'mw_additive' \n" + common_msg
            )
        if (
            self.config.is_isothermal
            and self.config.energy_balance_type != EnergyBalanceType.none
        ):
            raise ConfigurationError(
                "If the isothermal assumption is used then the energy balance type must be none"
            )

    def build(self):
        # build always starts by calling super().build()
        # This triggers a lot of boilerplate in the background for you
        super().build()

        # this creates blank scaling factors, which are populated later
        self.scaling_factor = Suffix(direction=Suffix.EXPORT)

        # Next, get the base units of measurement from the property definition
        units_meta = self.config.property_package.get_metadata().get_derived_units

        # Check configs for errors
        self._validate_config()

        # Assign name IDs locally for reference later when building constraints
        self.boron_name_id = self.config.chemical_mapping_data["boron_name"]
        self.borate_name_id = self.config.chemical_mapping_data["borate_name"]
        if "proton_name" in self.config.chemical_mapping_data:
            self.proton_name_id = self.config.chemical_mapping_data["proton_name"]
        else:
            self.proton_name_id = None
        if "hydroxide_name" in self.config.chemical_mapping_data:
            self.hydroxide_name_id = self.config.chemical_mapping_data["hydroxide_name"]
        else:
            self.hydroxide_name_id = None
        if "cation_name" in self.config.chemical_mapping_data["caustic_additive"]:
            self.cation_name_id = self.config.chemical_mapping_data["caustic_additive"][
                "cation_name"
            ]
        else:
            self.cation_name_id = None
        if "additive_name" in self.config.chemical_mapping_data["caustic_additive"]:
            self.caustic_chem_name = self.config.chemical_mapping_data[
                "caustic_additive"
            ]["additive_name"]
        else:
            self.caustic_chem_name = None

        # Cross reference and check given names with set of valid names
        if self.boron_name_id not in self.config.property_package.component_list:
            raise ConfigurationError(
                "\n Given 'boron_name' {"
                + self.boron_name_id
                + "} does not match "
                + "any species name from the property package \n{}".format(
                    [c for c in self.config.property_package.component_list]
                )
            )
        if self.borate_name_id not in self.config.property_package.component_list:
            raise ConfigurationError(
                "\n Given 'borate_name' {"
                + self.borate_name_id
                + "} does not match "
                + "any species name from the property package \n{}".format(
                    [c for c in self.config.property_package.component_list]
                )
            )
        if self.proton_name_id != None:
            if self.proton_name_id not in self.config.property_package.component_list:
                raise ConfigurationError(
                    "\n Given 'proton_name' {"
                    + self.proton_name_id
                    + "} does not match "
                    + "any species name from the property package \n{}".format(
                        [c for c in self.config.property_package.component_list]
                    )
                )
        if self.hydroxide_name_id != None:
            if (
                self.hydroxide_name_id
                not in self.config.property_package.component_list
            ):
                raise ConfigurationError(
                    "\n Given 'hydroxide_name' {"
                    + self.hydroxide_name_id
                    + "} does not match "
                    + "any species name from the property package \n{}".format(
                        [c for c in self.config.property_package.component_list]
                    )
                )
        if self.cation_name_id != None:
            if self.cation_name_id not in self.config.property_package.component_list:
                raise ConfigurationError(
                    "\n Given 'cation_name' {"
                    + self.cation_name_id
                    + "} does not match "
                    + "any species name from the property package \n{}".format(
                        [c for c in self.config.property_package.component_list]
                    )
                )

        # check for existence of inherent reactions
        #   This is to ensure that no degeneracy could be introduced
        #   in the system of equations (may not need this explicit check)
        if hasattr(self.config.property_package, "inherent_reaction_idx"):
            raise ConfigurationError(
                "\n Property Package CANNOT contain 'inherent_reactions' \n"
            )

        # cation set reference
        cation_set = self.config.property_package.cation_set

        # anion set reference
        anion_set = self.config.property_package.anion_set

        # Add param to store all charges of ions for convenience
        self.ion_charge = Param(
            anion_set | cation_set,
            initialize=1,
            mutable=True,
            units=pyunits.dimensionless,
            doc="Ion charge",
        )

        # Loop through full set and try to assign charge
        for j in self.config.property_package.component_list:
            if j in anion_set or j in cation_set:
                self.ion_charge[j] = self.config.property_package.get_component(
                    j
                ).config.charge

        # Add unit variables and parameters
        mw_add = pyunits.convert_value(
            self.config.chemical_mapping_data["caustic_additive"]["mw_additive"][0],
            from_units=self.config.chemical_mapping_data["caustic_additive"][
                "mw_additive"
            ][1],
            to_units=pyunits.kg / pyunits.mol,
        )
        self.caustic_mw = Param(
            mutable=True,
            initialize=mw_add,
            domain=NonNegativeReals,
            units=pyunits.kg / pyunits.mol,
            doc="Molecular weight of the caustic additive",
        )
        self.additive_molar_ratio = Param(
            mutable=True,
            initialize=self.config.chemical_mapping_data["caustic_additive"][
                "moles_cation_per_additive"
            ],
            domain=NonNegativeReals,
            units=pyunits.dimensionless,
            doc="Moles of cation per moles of caustic additive",
        )
        self.caustic_dose_rate = Var(
            self.flowsheet().config.time,
            initialize=0,
            bounds=(0, None),
            domain=NonNegativeReals,
            units=pyunits.kg / pyunits.s,
            doc="Dosage rate of the set of caustic additive",
        )

        # Reaction parameters
        self.Kw_0 = Param(
            mutable=True,
            initialize=60.91,
            domain=NonNegativeReals,
            units=pyunits.mol**2 / pyunits.m**6,
            doc="Water dissociation pre-exponential constant",
        )
        self.dH_w = Param(
            mutable=True,
            initialize=55830,
            domain=NonNegativeReals,
            units=pyunits.J / pyunits.mol,
            doc="Water dissociation enthalpy",
        )
        self.Ka_0 = Param(
            mutable=True,
            initialize=0.000163,
            domain=NonNegativeReals,
            units=pyunits.mol / pyunits.m**3,
            doc="Boron dissociation pre-exponential constant",
        )
        self.dH_a = Param(
            mutable=True,
            initialize=13830,
            domain=NonNegativeReals,
            units=pyunits.J / pyunits.mol,
            doc="Boron dissociation enthalpy",
        )

        # molarity vars (for approximate boron speciation)
        #       Used to establish the mass transfer rates by
        #       first solving a coupled equilibrium system
        #
        #   ENE: [H+] = [OH-] + [A-] + (Alk - n*[base])
        #   MB:  TB = [HA] + [A-]
        #   rw:  Kw = [H+][OH-]
        #   ra:  Ka[HA] = [H+][A-]
        #
        #       Alk = sum(n*Anions) - sum(n*Cations) (from props)
        #       [base] = (Dose/MW)
        #       TB = [HA]_inlet + [A-]_inlet (from props)

        # NOTE: These variables are internal to the unit model
        #   and are used to establish what the mass transfer
        #   constraints need to be in order to achieve a specific
        #   pH and boron speciation at the exit of the unit.
        self.conc_mol_H = Var(
            self.flowsheet().config.time,
            initialize=1e-4,
            bounds=(0, None),
            domain=NonNegativeReals,
            units=pyunits.mol / pyunits.m**3,
            doc="Resulting molarity of protons",
        )
        self.conc_mol_OH = Var(
            self.flowsheet().config.time,
            initialize=1e-4,
            bounds=(0, None),
            domain=NonNegativeReals,
            units=pyunits.mol / pyunits.m**3,
            doc="Resulting molarity of hydroxide",
        )
        self.conc_mol_Boron = Var(
            self.flowsheet().config.time,
            initialize=1e-2,
            bounds=(0, None),
            domain=NonNegativeReals,
            units=pyunits.mol / pyunits.m**3,
            doc="Resulting molarity of Boron",
        )
        self.conc_mol_Borate = Var(
            self.flowsheet().config.time,
            initialize=1e-2,
            bounds=(0, None),
            domain=NonNegativeReals,
            units=pyunits.mol / pyunits.m**3,
            doc="Resulting molarity of Borate",
        )

        # Variables for volume and retention time
        self.reactor_volume = Var(
            initialize=1,
            bounds=(0, None),
            domain=NonNegativeReals,
            units=pyunits.m**3,
            doc="Volume of the reactor",
        )
        self.reactor_retention_time = Var(
            self.flowsheet().config.time,
            initialize=500,
            bounds=(0, None),
            domain=NonNegativeReals,
            units=pyunits.s,
            doc="Hydraulic retention time of the reactor",
        )

        # Build control volume for feed side
        self.control_volume = ControlVolume0DBlock(
            dynamic=False,
            has_holdup=False,
            property_package=self.config.property_package,
            property_package_args=self.config.property_package_args,
            reaction_package=None,
            reaction_package_args=None,
        )

        self.control_volume.add_state_blocks(has_phase_equilibrium=False)

        self.control_volume.add_material_balances(
            balance_type=self.config.material_balance_type,
            has_mass_transfer=True,
            has_rate_reactions=False,
            has_equilibrium_reactions=False,
        )

        self.control_volume.add_energy_balances(
            balance_type=self.config.energy_balance_type,
            has_enthalpy_transfer=False,
        )

        if self.config.is_isothermal:
            self.control_volume.add_isothermal_assumption()

        self.control_volume.add_momentum_balances(
            balance_type=self.config.momentum_balance_type, has_pressure_change=False
        )

        # Add ports
        self.add_inlet_port(name="inlet", block=self.control_volume)
        self.add_outlet_port(name="outlet", block=self.control_volume)

        # Constraints for volume and retention time
        @self.Constraint(
            self.flowsheet().config.time,
            doc="Reactor volume constraint",
        )
        def eq_reactor_volume(self, t):
            Q = pyunits.convert(
                self.control_volume.properties_out[t].flow_vol_phase["Liq"],
                to_units=pyunits.m**3 / pyunits.s,
            )
            return self.reactor_volume == Q * self.reactor_retention_time[t]

        # Constraints for mass transfer terms
        @self.Constraint(
            self.flowsheet().config.time,
            doc="Electroneutrality condition",
        )
        def eq_electroneutrality(self, t):
            ResIons = 0
            for j in self.ion_charge:
                conc = self.control_volume.properties_out[t].conc_mol_phase_comp[
                    "Liq", j
                ]
                if (
                    j == self.boron_name_id
                    or j == self.borate_name_id
                    or j == self.proton_name_id
                    or j == self.hydroxide_name_id
                ):
                    ResIons += 0.0
                else:
                    ResIons += -self.ion_charge[j] * conc
            conc_mol_H = pyunits.convert(
                self.conc_mol_H[t],
                to_units=units_meta("amount") * units_meta("length") ** -3,
            )
            conc_mol_OH = pyunits.convert(
                self.conc_mol_OH[t],
                to_units=units_meta("amount") * units_meta("length") ** -3,
            )
            conc_mol_Borate = pyunits.convert(
                self.conc_mol_Borate[t],
                to_units=units_meta("amount") * units_meta("length") ** -3,
            )

            return conc_mol_H == conc_mol_OH + conc_mol_Borate + ResIons

        @self.Constraint(
            self.flowsheet().config.time,
            doc="Total boron balance",
        )
        def eq_total_boron(self, t):
            inlet_Boron = self.control_volume.properties_in[t].conc_mol_phase_comp[
                "Liq", self.boron_name_id
            ]
            inlet_Borate = self.control_volume.properties_in[t].conc_mol_phase_comp[
                "Liq", self.borate_name_id
            ]
            conc_mol_Borate = pyunits.convert(
                self.conc_mol_Borate[t],
                to_units=units_meta("amount") * units_meta("length") ** -3,
            )
            conc_mol_Boron = pyunits.convert(
                self.conc_mol_Boron[t],
                to_units=units_meta("amount") * units_meta("length") ** -3,
            )
            return inlet_Boron + inlet_Borate == conc_mol_Borate + conc_mol_Boron

        @self.Constraint(
            self.flowsheet().config.time,
            doc="Water dissociation",
        )
        def eq_water_dissociation(self, t):
            return (
                self.Kw_0
                * exp(
                    -self.dH_w
                    / Constants.gas_constant
                    / self.control_volume.properties_out[t].temperature
                )
            ) == self.conc_mol_H[t] * self.conc_mol_OH[t]

        @self.Constraint(
            self.flowsheet().config.time,
            doc="Boron dissociation",
        )
        def eq_boron_dissociation(self, t):
            return (
                self.Ka_0
                * exp(
                    -self.dH_a
                    / Constants.gas_constant
                    / self.control_volume.properties_out[t].temperature
                )
            ) * self.conc_mol_Boron[t] == self.conc_mol_H[t] * self.conc_mol_Borate[t]

        # Add constraints for mass transfer terms
        @self.Constraint(
            self.flowsheet().config.time,
            self.config.property_package.phase_list,
            self.config.property_package.component_list,
            doc="Mass transfer term",
        )
        def eq_mass_transfer_term(self, t, p, j):
            map = {
                self.boron_name_id: self.conc_mol_Boron[t],
                self.borate_name_id: self.conc_mol_Borate[t],
                self.proton_name_id: self.conc_mol_H[t],
                self.hydroxide_name_id: self.conc_mol_OH[t],
            }

            if (
                j == self.boron_name_id
                or j == self.borate_name_id
                or j == self.proton_name_id
                or j == self.hydroxide_name_id
            ):
                c_out = pyunits.convert(
                    map[j],
                    to_units=units_meta("amount") * units_meta("length") ** -3,
                )
                input_rate = self.control_volume.properties_in[t].flow_mol_phase_comp[
                    p, j
                ]
                exit_rate = (
                    self.control_volume.properties_out[t].flow_vol_phase[p] * c_out
                )

                loss_rate = input_rate - exit_rate
                return self.control_volume.mass_transfer_term[t, p, j] == -loss_rate

            elif j == self.cation_name_id:
                dose_rate = pyunits.convert(
                    self.caustic_dose_rate[t]
                    / self.caustic_mw
                    * self.additive_molar_ratio,
                    to_units=units_meta("amount") / units_meta("time"),
                )
                return self.control_volume.mass_transfer_term[t, p, j] == dose_rate

            else:
                return self.control_volume.mass_transfer_term[t, p, j] == 0.0

    # initialize method
    def initialize_build(
        self,
        state_args=None,
        outlvl=idaeslog.NOTSET,
        solver=None,
        optarg=None,
    ):
        """
        General wrapper for pressure changer initialization routines

        Keyword Arguments:
            state_args : a dict of arguments to be passed to the property
                         package(s) to provide an initial state for
                         initialization (see documentation of the specific
                         property package) (default = {}).
            outlvl : sets output level of initialization routine
            optarg : solver options dictionary object (default=None)
            solver : str indicating which solver to use during
                     initialization (default = None)

        Returns: None
        """
        init_log = idaeslog.getInitLogger(self.name, outlvl, tag="unit")
        solve_log = idaeslog.getSolveLogger(self.name, outlvl, tag="unit")
        # Set solver options
        opt = get_solver(solver, optarg)

        # ---------------------------------------------------------------------
        # Initialize holdup block
        flags = self.control_volume.initialize(
            outlvl=outlvl,
            optarg=optarg,
            solver=solver,
            state_args=state_args,
        )
        init_log.info_high("Initialization Step 1 Complete.")

        # Apply guess for unit model vars
        for t in self.flowsheet().config.time:
            # Naive guess (pH = 7)
            self.conc_mol_H[t].set_value(1e-4)
            self.conc_mol_OH[t].set_value(10**-14 * 1000 / self.conc_mol_H[t].value)
            TB = value(
                self.control_volume.properties_in[t].conc_mol_phase_comp[
                    "Liq", self.boron_name_id
                ]
            ) + value(
                self.control_volume.properties_in[t].conc_mol_phase_comp[
                    "Liq", self.borate_name_id
                ]
            )
            Ratio = 10**-9.21 * 1000 / self.conc_mol_H[t].value
            self.conc_mol_Boron[t].set_value(TB / (1 + Ratio))
            self.conc_mol_Borate[t].set_value(TB * Ratio / (1 + Ratio))
        # ---------------------------------------------------------------------

        # ---------------------------------------------------------------------
        # Solve unit
        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            res = opt.solve(self, tee=slc.tee)
        init_log.info_high("Initialization Step 2 {}.".format(idaeslog.condition(res)))

        # ---------------------------------------------------------------------
        # Release Inlet state
        self.control_volume.release_state(flags, outlvl + 1)
        init_log.info("Initialization Complete: {}".format(idaeslog.condition(res)))

        if not check_optimal_termination(res):
            raise InitializationError(f"Unit model {self.name} failed to initialize")

        # Rescale internal variables
        for t in self.flowsheet().config.time:
            iscale.set_scaling_factor(
                self.conc_mol_OH[t], max(100 / self.conc_mol_OH[t].value, 100)
            )
            iscale.set_scaling_factor(
                self.conc_mol_H[t], max(100 / self.conc_mol_H[t].value, 100)
            )
            iscale.set_scaling_factor(
                self.conc_mol_Boron[t], max(100 / self.conc_mol_Boron[t].value, 100)
            )
            iscale.set_scaling_factor(
                self.conc_mol_Borate[t], max(100 / self.conc_mol_Borate[t].value, 100)
            )

    def outlet_pH(self, time=0):
        return -log10(value(self.conc_mol_H[time]) / 1000)

    def outlet_pOH(self, time=0):
        return -log10(value(self.conc_mol_OH[time]) / 1000)

    def propogate_initial_state(self):
        units_meta = self.config.property_package.get_metadata().get_derived_units

        # This is a helper function to automate some of the logic behind
        #   setting good initial guesses for state vars and properties
        def _propogation_helper(
            name, prop_in, prop_out, is_indexed=False, func=None, obj=None
        ):
            if name in prop_in.define_state_vars():
                if is_indexed == False:
                    if prop_out.component(name).is_fixed() == False:
                        prop_out.component(name).set_value(
                            value(prop_in.component(name))
                        )
                else:
                    for ind in prop_in.component(name):
                        if prop_out.component(name)[ind].is_fixed() == False:
                            prop_out.component(name)[ind] = value(
                                prop_in.component(name)[ind]
                            )
            if (
                prop_out.is_property_constructed(name)
                and not name in prop_in.define_state_vars()
            ):
                if is_indexed == False:
                    if func == None and prop_in.is_property_constructed(name):
                        prop_out.component(name).set_value(
                            value(prop_in.component(name))
                        )
                    else:
                        prop_out.component(name).set_value(func(obj))
                        if prop_in.is_property_constructed(name):
                            prop_in.component(name).set_value(func(obj))
                else:
                    for ind in prop_out.component(name):
                        if func == None and prop_in.is_property_constructed(name):
                            prop_out.component(name)[ind] = value(
                                prop_in.component(name)[ind]
                            )
                        else:
                            prop_out.component(name)[ind] = func(obj, ind)
                            if prop_in.is_property_constructed(name):
                                prop_in.component(name)[ind] = func(obj, ind)

        t0 = self.flowsheet().time.first()
        for t in self.control_volume.properties_in:
            # Should check 'define_state_vars' to see if user has provided
            #   state vars that are outside of the checks in this function
            if (
                "flow_mol_phase_comp"
                not in self.control_volume.properties_in[t].define_state_vars()
                and "flow_mass_phase_comp"
                not in self.control_volume.properties_in[t].define_state_vars()
            ):
                raise ConfigurationError(
                    "BoronRemoval unit model requires "
                    "either a 'flow_mol_phase_comp' or 'flow_mass_phase_comp' "
                    "state variable basis to apply the 'propogate_initial_state' method"
                )

            _propogation_helper(
                "pressure",
                self.control_volume.properties_in[t],
                self.control_volume.properties_out[t],
                is_indexed=False,
            )

            _propogation_helper(
                "temperature",
                self.control_volume.properties_in[t],
                self.control_volume.properties_out[t],
                is_indexed=False,
            )

            _propogation_helper(
                "flow_mol_phase_comp",
                self.control_volume.properties_in[t],
                self.control_volume.properties_out[t],
                is_indexed=True,
            )

            def _flow_vol_calc(self, no_index):
                sum = 0
                for ind in self.control_volume.properties_in[t].flow_mol_phase_comp:
                    sum += self.control_volume.properties_in[t].flow_mol_phase_comp[ind]
                approx_dens = pyunits.convert(
                    55000 * pyunits.mol / pyunits.m**3,
                    to_units=units_meta("amount") * units_meta("length") ** -3,
                )
                return sum / approx_dens

            _propogation_helper(
                "flow_vol_phase",
                self.control_volume.properties_in[t],
                self.control_volume.properties_out[t],
                is_indexed=True,
                func=_flow_vol_calc,
                obj=self,
            )

            def _flow_mass_calc(self, ind):
                return value(
                    self.control_volume.properties_in[t0].flow_mol_phase_comp[ind]
                    * self.control_volume.properties_in[t0].mw_comp[ind[1]]
                )

            _propogation_helper(
                "flow_mass_phase_comp",
                self.control_volume.properties_in[t],
                self.control_volume.properties_out[t],
                is_indexed=True,
                func=_flow_mass_calc,
                obj=self,
            )

            def _mole_frac_calc(self, ind):
                return value(
                    self.control_volume.properties_in[t0].flow_mol_phase_comp[ind]
                    / sum(
                        self.control_volume.properties_in[t0].flow_mol_phase_comp[
                            ind[0], j
                        ]
                        for j in self.config.property_package.component_list
                    )
                )

            _propogation_helper(
                "mole_frac_phase_comp",
                self.control_volume.properties_in[t],
                self.control_volume.properties_out[t],
                is_indexed=True,
                func=_mole_frac_calc,
                obj=self,
            )

            def _mass_frac_calc(self, ind):
                return value(
                    self.control_volume.properties_in[t0].flow_mol_phase_comp[ind]
                    * self.control_volume.properties_in[t0].mw_comp[ind[1]]
                    / sum(
                        self.control_volume.properties_in[t0].flow_mol_phase_comp[
                            ind[0], j
                        ]
                        * self.control_volume.properties_in[t0].mw_comp[j]
                        for j in self.config.property_package.component_list
                    )
                )

            _propogation_helper(
                "mass_frac_phase_comp",
                self.control_volume.properties_in[t],
                self.control_volume.properties_out[t],
                is_indexed=True,
                func=_mass_frac_calc,
                obj=self,
            )

            def _conc_mol_calc(self, ind):
                approx_dens = pyunits.convert(
                    55000 * pyunits.mol / pyunits.m**3,
                    to_units=units_meta("amount") * units_meta("length") ** -3,
                )
                return approx_dens * value(
                    self.control_volume.properties_in[t0].flow_mol_phase_comp[ind]
                    / sum(
                        self.control_volume.properties_in[t0].flow_mol_phase_comp[
                            ind[0], j
                        ]
                        for j in self.config.property_package.component_list
                    )
                )

            _propogation_helper(
                "conc_mol_phase_comp",
                self.control_volume.properties_in[t],
                self.control_volume.properties_out[t],
                is_indexed=True,
                func=_conc_mol_calc,
                obj=self,
            )

            def _conc_mass_calc(self, ind):
                approx_dens = pyunits.convert(
                    1000 * pyunits.kg / pyunits.m**3,
                    to_units=units_meta("mass") * units_meta("length") ** -3,
                )
                return approx_dens * value(
                    self.control_volume.properties_in[t0].flow_mol_phase_comp[ind]
                    * self.control_volume.properties_in[t0].mw_comp[ind[1]]
                    / sum(
                        self.control_volume.properties_in[t0].flow_mol_phase_comp[
                            ind[0], j
                        ]
                        * self.control_volume.properties_in[t0].mw_comp[j]
                        for j in self.config.property_package.component_list
                    )
                )

            _propogation_helper(
                "conc_mass_phase_comp",
                self.control_volume.properties_in[t],
                self.control_volume.properties_out[t],
                is_indexed=True,
                func=_conc_mass_calc,
                obj=self,
            )

        # End Loop

    def calculate_scaling_factors(self):
        super().calculate_scaling_factors()

        units_meta = self.config.property_package.get_metadata().get_derived_units

        # Provide some intial guess values before scaling
        self.propogate_initial_state()

        # Add scaling for unit model vars (with user input)
        if iscale.get_scaling_factor(self.caustic_dose_rate) is None:
            sf = iscale.get_scaling_factor(
                self.caustic_dose_rate, default=1e4, warning=True
            )
            iscale.set_scaling_factor(self.caustic_dose_rate, sf)

        if iscale.get_scaling_factor(self.reactor_volume) is None:
            sf = iscale.get_scaling_factor(self.reactor_volume, default=1, warning=True)
            iscale.set_scaling_factor(self.reactor_volume, sf)

        if iscale.get_scaling_factor(self.reactor_retention_time) is None:
            sf = iscale.get_scaling_factor(
                self.reactor_retention_time, default=1e-2, warning=True
            )
            iscale.set_scaling_factor(self.reactor_retention_time, sf)

        # Add scaling for unit model vars (without user input)
        if iscale.get_scaling_factor(self.conc_mol_Boron) is None:
            sf = iscale.get_scaling_factor(
                self.control_volume.properties_in[0].conc_mol_phase_comp[
                    "Liq", self.boron_name_id
                ],
                default=1,
                warning=False,
            )
            iscale.set_scaling_factor(self.conc_mol_Boron, sf / 10)

        if iscale.get_scaling_factor(self.conc_mol_Borate) is None:
            sf = iscale.get_scaling_factor(
                self.control_volume.properties_in[0].conc_mol_phase_comp[
                    "Liq", self.borate_name_id
                ],
                default=1,
                warning=False,
            )
            iscale.set_scaling_factor(self.conc_mol_Borate, sf / 10)

        # Scaling for H and OH
        if iscale.get_scaling_factor(self.conc_mol_H) is None:
            if self.proton_name_id in self.config.property_package.component_list:
                sf = iscale.get_scaling_factor(
                    self.control_volume.properties_in[0].conc_mol_phase_comp[
                        "Liq", self.proton_name_id
                    ],
                    default=1,
                    warning=False,
                )
            else:
                sf = 10
            iscale.set_scaling_factor(self.conc_mol_H, sf)

        if iscale.get_scaling_factor(self.conc_mol_OH) is None:
            if self.hydroxide_name_id in self.config.property_package.component_list:
                sf = iscale.get_scaling_factor(
                    self.control_volume.properties_in[0].conc_mol_phase_comp[
                        "Liq", self.hydroxide_name_id
                    ],
                    default=1,
                    warning=False,
                )
            else:
                sf = 10
            iscale.set_scaling_factor(self.conc_mol_OH, sf)

        # Scale reactor volume constraint
        sf = iscale.get_scaling_factor(self.reactor_volume)
        for t in self.control_volume.properties_in:
            iscale.constraint_scaling_transform(self.eq_reactor_volume[t], sf)

        # Scaling for water dissociation and boron dissociation
        for t in self.control_volume.properties_in:
            sf = iscale.get_scaling_factor(self.conc_mol_H) * iscale.get_scaling_factor(
                self.conc_mol_OH
            )
            iscale.constraint_scaling_transform(self.eq_water_dissociation[t], sf)
        for t in self.control_volume.properties_in:
            sf = iscale.get_scaling_factor(
                self.conc_mol_Borate
            ) * iscale.get_scaling_factor(self.conc_mol_H)
            iscale.constraint_scaling_transform(self.eq_boron_dissociation[t], sf)

        # Scaling for total boron
        for t in self.control_volume.properties_in:
            sf = iscale.get_scaling_factor(self.conc_mol_Boron)
            iscale.constraint_scaling_transform(self.eq_total_boron[t], sf)

        # Scaling for electroneutrality
        for t in self.control_volume.properties_in:
            sf = iscale.get_scaling_factor(self.conc_mol_H) + iscale.get_scaling_factor(
                self.conc_mol_Borate
            )
            iscale.constraint_scaling_transform(self.eq_electroneutrality[t], sf)

        # Scaling for mass_transfer_term
        for t in self.control_volume.properties_in:
            for j in self.config.property_package.component_list:
                sf = iscale.get_scaling_factor(self.conc_mol_Borate)
                iscale.constraint_scaling_transform(
                    self.eq_mass_transfer_term[t, "Liq", j], sf
                )



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
import pytest

from pyomo.environ import ConcreteModel, assert_optimal_termination, value
from pyomo.util.check_units import assert_units_consistent
from idaes.core import FlowsheetBlock
from idaes.core.solvers import get_solver
from idaes.core.util.model_statistics import degrees_of_freedom
import idaes.core.util.scaling as iscale

from watertap.unit_models.mvc.components import Compressor
import watertap.property_models.water_prop_pack as props

solver = get_solver()

# -----------------------------------------------------------------------------
@pytest.mark.component
def test_compressor():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)
    m.fs.properties = props.WaterParameterBlock()
    m.fs.compressor = Compressor(property_package=m.fs.properties)
    # m.fs.compressor.control_volume.display()

    # scaling
    m.fs.properties.set_default_scaling("flow_mass_phase_comp", 1, index=("Vap", "H2O"))
    m.fs.properties.set_default_scaling("flow_mass_phase_comp", 1, index=("Liq", "H2O"))
    iscale.set_scaling_factor(m.fs.compressor.control_volume.work, 1e-6)
    iscale.calculate_scaling_factors(m)

    # state variables
    m.fs.compressor.inlet.flow_mass_phase_comp[0, "Vap", "H2O"].fix(1)
    m.fs.compressor.inlet.flow_mass_phase_comp[0, "Liq", "H2O"].fix(1e-8)
    m.fs.compressor.inlet.temperature[0].fix(350)  # K
    m.fs.compressor.inlet.pressure[0].fix(0.5e5)  # Pa

    # specifications
    m.fs.compressor.pressure_ratio.fix(2)
    m.fs.compressor.efficiency.fix(0.8)

    # check build
    assert_units_consistent(m)
    assert degrees_of_freedom(m) == 0

    # initialize
    m.fs.compressor.initialize()

    # solve
    solver = get_solver()
    results = solver.solve(m, tee=False)
    assert_optimal_termination(results)

    assert pytest.approx(1.1539e5, rel=1e-4) == value(
        m.fs.compressor.control_volume.work[0]
    )
    assert pytest.approx(1e-08, rel=1e-4) == value(
        m.fs.compressor.outlet.flow_mass_phase_comp[0, "Liq", "H2O"]
    )
    assert pytest.approx(1.0, rel=1e-4) == value(
        m.fs.compressor.outlet.flow_mass_phase_comp[0, "Vap", "H2O"]
    )
    assert pytest.approx(429.57, rel=1e-4) == value(
        m.fs.compressor.outlet.temperature[0]
    )
    assert pytest.approx(1.0e5, rel=1e-4) == value(m.fs.compressor.outlet.pressure[0])

    m.fs.compressor.report()
    perf_dict = m.fs.compressor._get_performance_contents()
    assert perf_dict == {
        "vars": {
            "Pressure ratio": m.fs.compressor.pressure_ratio,
            "Efficiency": m.fs.compressor.efficiency,
            "Work": m.fs.compressor.control_volume.work[0],
        }
    }



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################

import pyomo.environ as pyo
from ..util import (
    register_costing_parameter_block,
    make_capital_cost_var,
    make_fixed_operating_cost_var,
)


def build_electrolyzer_cost_param_block(blk):
    blk.factor_membrane_replacement = pyo.Var(
        initialize=0.33,
        doc="membrane replacement factor (fraction of membrane replaced/year)",
        units=pyo.units.year**-1,
    )
    blk.membrane_unit_cost = pyo.Var(
        initialize=25,  # (Yee, 2012)
        doc="membrane unit cost",
        units=pyo.units.USD_2012 / (pyo.units.meter**2),
    )
    blk.anode_unit_cost = pyo.Var(
        initialize=300,  # assumed
        doc="anode unit cost",
        units=pyo.units.USD_2005 / (pyo.units.meter**2),
    )
    blk.cathode_unit_cost = pyo.Var(
        initialize=600,  # (O'Brien, 2005)
        doc="cathode unit cost",
        units=pyo.units.USD_2005 / (pyo.units.meter**2),
    )
    blk.fraction_material_cost = pyo.Var(
        initialize=0.65,
        doc="membrane, anode, and cathode fraction of total capital",
        units=pyo.units.dimensionless,
    )


@register_costing_parameter_block(
    build_rule=build_electrolyzer_cost_param_block,
    parameter_block_name="electrolyzer",
)
def cost_electrolyzer(blk):
    """
    electrolyzer costing method
    """

    # ---------------------------------------------------------------------
    make_capital_cost_var(blk)
    blk.membrane_cost = pyo.Var(
        initialize=1e5,
        domain=pyo.NonNegativeReals,
        units=blk.costing_package.base_currency,
        doc="membrane capital cost",
    )
    blk.anode_cost = pyo.Var(
        initialize=1e5,
        domain=pyo.NonNegativeReals,
        units=blk.costing_package.base_currency,
        doc="anode capital cost",
    )
    blk.cathode_cost = pyo.Var(
        initialize=1e5,
        domain=pyo.NonNegativeReals,
        units=blk.costing_package.base_currency,
        doc="cathode capital cost",
    )
    blk.membrane_replacement_cost = pyo.Var(
        initialize=1e5,
        domain=pyo.NonNegativeReals,
        units=blk.costing_package.base_currency / blk.costing_package.base_period,
        doc="membrane replacement cost",
    )

    # ---------------------------------------------------------------------
    make_fixed_operating_cost_var(blk)

    # not using cost_membrane() utility to bypass required naming conventions
    blk.membrane_cost_constraint = pyo.Constraint(
        expr=blk.membrane_cost
        == pyo.units.convert(
            blk.costing_package.electrolyzer.membrane_unit_cost
            * blk.unit_model.membrane_area,
            to_units=blk.costing_package.base_currency,
        )
    )
    blk.anode_cost_constraint = pyo.Constraint(
        expr=blk.anode_cost
        == pyo.units.convert(
            blk.costing_package.electrolyzer.anode_unit_cost
            * blk.unit_model.anode_area,
            to_units=blk.costing_package.base_currency,
        )
    )
    blk.cathode_cost_constraint = pyo.Constraint(
        expr=blk.cathode_cost
        == pyo.units.convert(
            blk.costing_package.electrolyzer.cathode_unit_cost
            * blk.unit_model.cathode_area,
            to_units=blk.costing_package.base_currency,
        )
    )
    blk.costing_package.add_cost_factor(blk, "TIC")
    blk.capital_cost_constraint = pyo.Constraint(
        expr=blk.capital_cost
        == blk.cost_factor
        * (blk.membrane_cost + blk.anode_cost + blk.cathode_cost)
        / blk.costing_package.electrolyzer.fraction_material_cost
    )

    # ---------------------------------------------------------------------

    blk.membrane_replacement_cost_constraint = pyo.Constraint(
        expr=blk.membrane_replacement_cost
        == pyo.units.convert(
            blk.costing_package.electrolyzer.factor_membrane_replacement
            * blk.costing_package.electrolyzer.membrane_unit_cost
            * blk.unit_model.membrane_area,
            to_units=blk.costing_package.base_currency
            / blk.costing_package.base_period,
        )
    )
    blk.costing_package.cost_flow(
        pyo.units.convert(blk.unit_model.power, to_units=pyo.units.kW),
        "electricity",
    )
    blk.fixed_operating_cost_constraint = pyo.Constraint(
        expr=blk.fixed_operating_cost == blk.membrane_replacement_cost
    )



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################

"""
Tests for zero-order peracetic acid disinfection model
"""

import pytest, os

from pyomo.environ import (
    Block,
    Var,
    Constraint,
    ConcreteModel,
    value,
    assert_optimal_termination,
)
from pyomo.util.check_units import assert_units_consistent

from idaes.core import FlowsheetBlock
from idaes.core.solvers import get_solver
from idaes.core.util.model_statistics import degrees_of_freedom
from idaes.core.util.testing import initialization_tester
from idaes.core import UnitModelCostingBlock

from watertap.unit_models.zero_order import PeraceticAcidDisinfectionZO
from watertap.core.wt_database import Database
from watertap.core.zero_order_properties import WaterParameterBlock
from watertap.costing.zero_order_costing import ZeroOrderCosting

solver = get_solver()


class TestPeraceticAcidDisinfection:
    @pytest.fixture(scope="class")
    def model(self):
        m = ConcreteModel()
        m.db = Database()

        m.fs = FlowsheetBlock(dynamic=False)
        m.fs.params = WaterParameterBlock(
            solute_list=["peracetic_acid", "total_coliforms_fecal_ecoli"]
        )

        m.fs.unit = PeraceticAcidDisinfectionZO(
            property_package=m.fs.params, database=m.db
        )

        # Inlet mass flowrates in kg/s
        m.fs.unit.inlet.flow_mass_comp[0, "H2O"].fix(2800)
        m.fs.unit.inlet.flow_mass_comp[0, "peracetic_acid"].fix(0.005)
        m.fs.unit.inlet.flow_mass_comp[0, "total_coliforms_fecal_ecoli"].fix(5.56e-7)

        return m

    @pytest.mark.unit
    def test_build(self, model):
        assert model.fs.unit.config.database is model.db
        assert model.fs.unit._tech_type == "peracetic_acid_disinfection"
        assert isinstance(model.fs.unit.electricity, Var)
        assert isinstance(model.fs.unit.energy_electric_flow_vol_inlet, Var)
        assert isinstance(model.fs.unit.HRT, Var)
        assert isinstance(model.fs.unit.ecoli_cell_mass, Var)
        assert isinstance(model.fs.unit.disinfection_solution_wt_frac_PAA, Var)
        assert isinstance(model.fs.unit.disinfection_solution_density, Var)
        assert isinstance(model.fs.unit.disinfection_solution_flow_vol, Var)
        assert isinstance(model.fs.unit.disinfection_solution_flow_vol_rule, Constraint)
        assert isinstance(model.fs.unit.reactor_volume, Var)
        assert isinstance(model.fs.unit.reactor_volume_rule, Constraint)
        assert isinstance(model.fs.unit.inlet_ecoli_conc, Var)
        assert isinstance(model.fs.unit.outlet_ecoli_conc, Var)
        assert isinstance(model.fs.unit.ecoli_inlet_concentration_rule, Constraint)
        assert isinstance(model.fs.unit.ecoli_outlet_concentration_rule, Constraint)

    def test_load_parameters(self, model):
        data = model.db.get_unit_operation_parameters("peracetic_acid_disinfection")
        model.fs.unit.load_parameters_from_database(use_default_removal=True)

        assert model.fs.unit.recovery_frac_mass_H2O[0].fixed
        assert (
            model.fs.unit.recovery_frac_mass_H2O[0].value
            == data["recovery_frac_mass_H2O"]["value"]
        )

        for (t, j), v in model.fs.unit.removal_frac_mass_comp.items():
            assert v.fixed
            if j not in data["removal_frac_mass_comp"].keys():
                assert v.value == data["default_removal_frac_mass_comp"]["value"]
            else:
                assert v.value == data["removal_frac_mass_comp"][j]["value"]

        assert model.fs.unit.HRT.fixed
        assert model.fs.unit.HRT.value == data["HRT"]["value"]

        assert model.fs.unit.ecoli_cell_mass.fixed
        assert model.fs.unit.ecoli_cell_mass.value == data["ecoli_cell_mass"]["value"]

        assert model.fs.unit.disinfection_solution_wt_frac_PAA.fixed
        assert (
            model.fs.unit.disinfection_solution_wt_frac_PAA.value
            == data["disinfection_solution_wt_frac_PAA"]["value"]
        )

        assert model.fs.unit.disinfection_solution_density.fixed
        assert (
            model.fs.unit.disinfection_solution_density.value
            == data["disinfection_solution_density"]["value"]
        )

    @pytest.mark.component
    def test_degrees_of_freedom(self, model):
        assert degrees_of_freedom(model.fs.unit) == 0

    @pytest.mark.component
    def test_unit_consistency(self, model):
        assert_units_consistent(model.fs.unit)

    @pytest.mark.component
    def test_initialize(self, model):
        initialization_tester(model)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solve(self, model):
        results = solver.solve(model)

        # Check for optimal solution
        assert_optimal_termination(results)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solution(self, model):
        assert pytest.approx(6552, rel=1e-3) == value(
            model.fs.unit.reactor_volume  # m3
        )
        assert pytest.approx(0.029369, rel=1e-3) == value(
            model.fs.unit.disinfection_solution_flow_vol[0]  # L/s
        )

        assert pytest.approx(2.8, rel=1e-3) == value(
            model.fs.unit.properties_in[0].flow_vol  # m3/s
        )
        assert pytest.approx(0.0017857, rel=1e-3) == value(
            model.fs.unit.properties_in[0].conc_mass_comp["peracetic_acid"]  # kg/m3
        )
        assert pytest.approx(1.98571e-7, rel=1e-3) == value(
            model.fs.unit.properties_in[0].conc_mass_comp[
                "total_coliforms_fecal_ecoli"
            ]  # kg/m3
        )
        assert pytest.approx(198571, rel=1e-3) == value(
            model.fs.unit.inlet_ecoli_conc[0]  # 1/L
        )

        assert pytest.approx(2.8, rel=1e-3) == value(
            model.fs.unit.properties_treated[0].flow_vol  # m3/s
        )
        assert pytest.approx(0.0004464, rel=1e-3) == value(
            model.fs.unit.properties_treated[0].conc_mass_comp[
                "peracetic_acid"
            ]  # kg/m3
        )
        assert pytest.approx(9.92857e-10, rel=1e-3) == value(
            model.fs.unit.properties_treated[0].conc_mass_comp[
                "total_coliforms_fecal_ecoli"
            ]  # kg/m3
        )
        assert pytest.approx(992.857, rel=1e-3) == value(
            model.fs.unit.outlet_ecoli_conc[0]  # 1/L
        )

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_conservation(self, model):
        for j in model.fs.params.component_list:
            assert 1e-6 >= abs(
                value(
                    model.fs.unit.inlet.flow_mass_comp[0, j]
                    + sum(
                        model.fs.unit.generation_rxn_comp[0, r, j]
                        for r in model.fs.unit.reaction_set
                    )
                    - model.fs.unit.treated.flow_mass_comp[0, j]
                    - model.fs.unit.byproduct.flow_mass_comp[0, j]
                )
            )

    @pytest.mark.component
    def test_report(self, model):
        model.fs.unit.report()


def test_costing():
    m = ConcreteModel()
    m.db = Database()
    m.fs = FlowsheetBlock(dynamic=False)
    m.fs.params = WaterParameterBlock(
        solute_list=["peracetic_acid", "total_coliforms_fecal_ecoli"]
    )
    source_file = os.path.join(
        os.path.dirname(os.path.abspath(__file__)),
        "..",
        "..",
        "..",
        "examples",
        "flowsheets",
        "case_studies",
        "wastewater_resource_recovery",
        "peracetic_acid_disinfection",
        "peracetic_acid_case_study.yaml",
    )
    m.fs.costing = ZeroOrderCosting(case_study_definition=source_file)
    m.fs.unit = PeraceticAcidDisinfectionZO(property_package=m.fs.params, database=m.db)

    # Inlet mass flowrates in kg/s
    m.fs.unit.inlet.flow_mass_comp[0, "H2O"].fix(2800)
    m.fs.unit.inlet.flow_mass_comp[0, "peracetic_acid"].fix(0.005)
    m.fs.unit.inlet.flow_mass_comp[0, "total_coliforms_fecal_ecoli"].fix(5.56e-7)

    m.fs.unit.load_parameters_from_database(use_default_removal=True)

    assert degrees_of_freedom(m.fs.unit) == 0

    m.fs.unit.costing = UnitModelCostingBlock(flowsheet_costing_block=m.fs.costing)

    assert isinstance(m.fs.costing.peracetic_acid_disinfection, Block)
    assert isinstance(m.fs.costing.peracetic_acid_disinfection.sizing_cost, Var)

    assert isinstance(m.fs.unit.costing.capital_cost, Var)
    assert isinstance(m.fs.unit.costing.capital_cost_constraint, Constraint)

    assert_units_consistent(m.fs)
    assert degrees_of_freedom(m.fs.unit) == 0
    initialization_tester(m)
    results = solver.solve(m)
    assert_optimal_termination(results)

    assert pytest.approx(9676817.28, rel=1e-3) == value(m.fs.unit.costing.capital_cost)

    assert m.fs.unit.electricity[0] in m.fs.costing._registered_flows["electricity"]
    assert (
        m.fs.unit.disinfection_solution_flow_vol[0]
        in m.fs.costing._registered_flows["disinfection_solution"]
    )



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################

"""Pretreatment flowsheet components"""

from pyomo.environ import ConcreteModel, TransformationFactory
from pyomo.network import Arc
from idaes.core import FlowsheetBlock
from idaes.models.unit_models import (
    Separator,
    Mixer,
)
from idaes.models.unit_models.separator import (
    SplittingType,
    EnergySplittingType,
)
from idaes.core.util.scaling import (
    calculate_scaling_factors,
    set_scaling_factor,
    get_scaling_factor,
    constraint_scaling_transform,
)
from idaes.core.util.initialization import propagate_state
from watertap.unit_models.pressure_changer import Pump
from watertap.examples.flowsheets.full_treatment_train.flowsheet_components import (
    feed_block,
)
from watertap.examples.flowsheets.full_treatment_train.model_components import (
    unit_separator,
    unit_ZONF,
    property_models,
)
from watertap.examples.flowsheets.full_treatment_train.util import (
    solve_block,
    check_dof,
)


def build_pretreatment_NF(m, has_bypass=True, NF_type="ZO", NF_base="ion"):
    """
    Builds NF pretreatment including specified feed and auxiliary equipment.
    Arguments:
        has_bypass: True or False, default = True
        NF_type: 'Sep' or 'ZO', default = 'ZO'
        NF_base: 'ion' or 'salt', default = 'ion'
    """
    pretrt_port = {}
    prop = property_models.get_prop(m, base=NF_base)

    # build feed
    feed_block.build_feed(m, base=NF_base)

    # build NF
    if NF_type == "Sep":
        unit_separator.build_SepNF(m, base=NF_base)
    elif NF_type == "ZO":
        unit_ZONF.build_ZONF(m, base=NF_base)
        m.fs.pump_NF = Pump(property_package=prop)
    else:
        raise ValueError(
            "Unexpected model type {NF_type} provided to build_NF_no_bypass"
            "".format(NF_type=NF_type)
        )

    if has_bypass:
        # build auxiliary units
        m.fs.splitter = Separator(
            property_package=prop,
            outlet_list=["pretreatment", "bypass"],
            split_basis=SplittingType.totalFlow,
            energy_split_basis=EnergySplittingType.equal_temperature,
        )
        m.fs.mixer = Mixer(property_package=prop, inlet_list=["pretreatment", "bypass"])

        # connect models
        m.fs.s_pretrt_feed_splitter = Arc(
            source=m.fs.feed.outlet, destination=m.fs.splitter.inlet
        )
        m.fs.s_pretrt_splitter_mixer = Arc(
            source=m.fs.splitter.bypass, destination=m.fs.mixer.bypass
        )
        if NF_type == "ZO":
            m.fs.s_pretrt_splitter_pumpNF = Arc(
                source=m.fs.splitter.pretreatment, destination=m.fs.pump_NF.inlet
            )
            m.fs.s_pretrt_pumpNF_NF = Arc(
                source=m.fs.pump_NF.outlet, destination=m.fs.NF.inlet
            )
        else:
            m.fs.s_pretrt_splitter_NF = Arc(
                source=m.fs.splitter.pretreatment, destination=m.fs.NF.inlet
            )
        m.fs.s_pretrt_NF_mixer = Arc(
            source=m.fs.NF.permeate, destination=m.fs.mixer.pretreatment
        )

        # specify (NF and feed is already specified, mixer has 0 DOF, splitter has 1 DOF, NF pump has 2 DOF)
        # splitter
        m.fs.splitter.split_fraction[0, "bypass"].fix(0.1)
        if NF_type == "ZO":
            m.fs.pump_NF.efficiency_pump.fix(0.80)
            m.fs.pump_NF.control_volume.properties_out[0].pressure.fix(4e5)

        # inlet/outlet ports for pretreatment
        pretrt_port["out"] = m.fs.mixer.outlet
        pretrt_port["waste"] = m.fs.NF.retentate

    else:  # no bypass
        # build auxiliary units (none)

        # connect models
        if NF_type == "ZO":
            m.fs.s_pretrt_feed_pumpNF = Arc(
                source=m.fs.feed.outlet, destination=m.fs.pump_NF.inlet
            )
            m.fs.s_pretrt_pumpNF_NF = Arc(
                source=m.fs.pump_NF.outlet, destination=m.fs.NF.inlet
            )
            # TODO: should source be m.fs.pump_NF.outlet? Double-check here and other arcs with pump_NF
        else:
            m.fs.s_pretrt_feed_NF = Arc(
                source=m.fs.feed.outlet, destination=m.fs.NF.inlet
            )

        # specify (NF and feed are already specified, NF pump has 2 DOF)
        if NF_type == "ZO":
            m.fs.pump_NF.efficiency_pump.fix(0.80)
            m.fs.pump_NF.control_volume.properties_out[0].pressure.fix(4e5)

        # inlet/outlet ports for pretreatment
        pretrt_port["out"] = m.fs.NF.permeate
        pretrt_port["waste"] = m.fs.NF.retentate

    return pretrt_port


def scale_pretreatment_NF(m, **kwargs):
    calculate_scaling_factors(m.fs.feed)
    calculate_scaling_factors(m.fs.NF)

    if kwargs["has_bypass"]:
        calculate_scaling_factors(m.fs.splitter)
        set_scaling_factor(
            m.fs.splitter.split_fraction, 1
        )  # TODO: should have an IDAES default
        constraint_scaling_transform(
            m.fs.splitter.sum_split_frac[0], 1
        )  # TODO: should have an IDAES default
        calculate_scaling_factors(m.fs.mixer)
        set_scaling_factor(
            m.fs.mixer.minimum_pressure,
            get_scaling_factor(m.fs.mixer.mixed_state[0].pressure),
        )  # TODO: IDAES should have a default and link to the constraint
        for c in [
            m.fs.mixer.minimum_pressure_constraint[0, 1],
            m.fs.mixer.minimum_pressure_constraint[0, 2],
            m.fs.mixer.mixture_pressure[0.0],
        ]:
            constraint_scaling_transform(
                c, get_scaling_factor(m.fs.mixer.minimum_pressure)
            )

    if kwargs["NF_type"] == "ZO":
        set_scaling_factor(m.fs.pump_NF.control_volume.work, 1e-3)
        calculate_scaling_factors(m.fs.pump_NF)
        set_scaling_factor(
            m.fs.pump_NF.ratioP, 1
        )  # TODO: IDAES should have a default and link to the constraint


def initialize_pretreatment_NF(m, **kwargs):
    optarg = {"nlp_scaling_method": "user-scaling"}

    if kwargs["has_bypass"]:
        m.fs.feed.initialize(optarg=optarg)
        propagate_state(m.fs.s_pretrt_feed_splitter)
        m.fs.splitter.initialize(optarg=optarg)
        propagate_state(m.fs.s_pretrt_splitter_mixer)
        if kwargs["NF_type"] == "ZO":
            propagate_state(m.fs.s_pretrt_splitter_pumpNF)
            m.fs.pump_NF.initialize(optarg=optarg)
            propagate_state(m.fs.s_pretrt_pumpNF_NF)
            m.fs.NF.initialize(optarg=optarg)
        else:  # NF_type == 'Sep'
            propagate_state(m.fs.s_pretrt_splitter_NF)
            # m.fs.NF.initialize(optarg=optarg)  # IDAES error when NF is a separator TODO: address in IDAES
        propagate_state(m.fs.s_pretrt_NF_mixer)
        m.fs.mixer.initialize(optarg=optarg)

    else:  # no bypass
        m.fs.feed.initialize(optarg=optarg)
        if kwargs["NF_type"] == "ZO":
            propagate_state(m.fs.s_pretrt_feed_pumpNF)
            m.fs.pump_NF.initialize(optarg=optarg)
            propagate_state(m.fs.s_pretrt_pumpNF_NF)
            m.fs.NF.initialize(optarg=optarg)
        else:  # NF_type == 'Sep'
            propagate_state(m.fs.s_pretrt_feed_NF)
            # m.fs.NF.initialize(optarg=optarg)  # IDAES error when NF is a separator TODO: address in IDAES


def display_pretreatment_NF(m, **kwargs):

    m.fs.feed.report()

    if kwargs["has_bypass"]:
        m.fs.splitter.report()
        if kwargs["NF_type"] == "ZO":
            m.fs.pump_NF.report()
        m.fs.NF.report()
        m.fs.mixer.report()
    else:  # no bypass
        if kwargs["NF_type"] == "ZO":
            m.fs.pump_NF.report()
        m.fs.NF.report()


def solve_pretreatment_NF(**kwargs):
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)

    property_models.build_prop(m, base=kwargs["NF_base"])
    build_pretreatment_NF(m, **kwargs)
    TransformationFactory("network.expand_arcs").apply_to(m)

    scale_pretreatment_NF(m, **kwargs)

    initialize_pretreatment_NF(m, **kwargs)

    check_dof(m)
    solve_block(m, tee=True, fail_flag=True)

    display_pretreatment_NF(m, **kwargs)


if __name__ == "__main__":
    solve_pretreatment_NF(has_bypass=True, NF_type="ZO", NF_base="ion")



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
from watertap.tools.parameter_sweep import (
    LinearSample,
    parameter_sweep,
)
import watertap.examples.flowsheets.case_studies.wastewater_resource_recovery.swine_wwt.swine_wwt as swine_wwt


def set_up_sensitivity(m):
    outputs = {}
    optimize_kwargs = {"fail_flag": False}
    opt_function = swine_wwt.solve

    # create outputs
    outputs["LCOW"] = m.fs.costing.levelized_costs.LCOW
    outputs["LCOT"] = m.fs.costing.levelized_costs.LCOT
    outputs["LCOH2"] = m.fs.costing.levelized_costs.LCOH2
    outputs["LCOCOD"] = m.fs.costing.levelized_costs.LCOCOD
    outputs["LCOP"] = m.fs.costing.levelized_costs.LCOP
    outputs["LCOVFA"] = m.fs.costing.levelized_costs.LCOVFA
    outputs["LCON"] = m.fs.costing.levelized_costs.LCON

    return outputs, optimize_kwargs, opt_function


def run_analysis(case_num=1, nx=5, interpolate_nan_outputs=True, save_outputs=None):
    case_num = int(case_num)
    nx = int(nx)
    interpolate_nan_outputs = bool(interpolate_nan_outputs)

    m = swine_wwt.main()[0]

    outputs, optimize_kwargs, opt_function = set_up_sensitivity(m)

    sweep_params = {}
    if case_num == 1:
        # AnMBR-MEC Unit Capex and Opex
        sweep_params["AnMBR-MEC Unit Capex"] = LinearSample(
            m.fs.costing.anaerobic_mbr_mec.unit_capex, 0.08, 8, nx
        )
        sweep_params["AnMBR-MEC Unit Opex"] = LinearSample(
            m.fs.costing.anaerobic_mbr_mec.unit_capex, 0.006, 6, nx
        )
    elif case_num == 2:
        # AnMBR-MEC Water Recovery
        sweep_params["AnMBR-MEC H2O Recovery"] = LinearSample(
            m.fs.mbr_mec.recovery_frac_mass_H2O, 0.1, 1, nx
        )
        # AnMBR-MEC Conversion to ffCOD
        sweep_params["AnMBR-MEC Conversion to ffCOD"] = LinearSample(
            m.fs.mbr_mec.reaction_conversion[0, "cod_to_nonbiodegradable_cod"],
            0.3,
            0.7,
            nx,
        )
    elif case_num == 3:
        # VFA Recovery Unit Capex and heating costs
        sweep_params["VFA Rec Unit Capex"] = LinearSample(
            m.fs.costing.vfa_recovery.unit_capex, 0.3, 30, nx
        )
        sweep_params["Cost of Heat"] = LinearSample(m.fs.costing.heat_cost, 0, 0.15, nx)
    else:
        raise ValueError(f"{case_num} is not yet implemented")

    global_results = parameter_sweep(
        m,
        sweep_params,
        outputs,
        csv_results_file_name=save_outputs,
        optimize_function=opt_function,
        optimize_kwargs=optimize_kwargs,
        interpolate_nan_outputs=interpolate_nan_outputs,
    )

    return global_results, sweep_params, m


if __name__ == "__main__":
    results, sweep_params, m = run_analysis()



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
"""
This module contains a zero-order representation of a Chlorination unit.
"""

import pyomo.environ as pyo
from pyomo.environ import units as pyunits, Var
from idaes.core import declare_process_block_class
from watertap.core import build_siso, constant_intensity, ZeroOrderBaseData

# Some more information about this module
__author__ = "Kurban Sitterley"


@declare_process_block_class("ChlorinationZO")
class ChlorinationZOData(ZeroOrderBaseData):
    """
    Zero-Order model for a Chlorination unit operation.
    """

    CONFIG = ZeroOrderBaseData.CONFIG()

    def build(self):
        super().build()

        self._tech_type = "chlorination"

        build_siso(self)
        constant_intensity(self)

        self.initial_chlorine_demand = Var(
            self.flowsheet().time,
            units=pyunits.mg / pyunits.liter,
            doc="Initial chlorine demand",
        )

        self.contact_time = Var(
            self.flowsheet().time, units=pyunits.hour, doc="Chlorine contact time"
        )
        self.concentration_time = Var(
            self.flowsheet().time,
            units=(pyunits.mg * pyunits.minute) / pyunits.liter,
            doc="CT value for chlorination",
        )
        self.chlorine_decay_rate = Var(
            self.flowsheet().time,
            units=pyunits.mg / (pyunits.L * pyunits.hour),
            doc="Chlorine decay rate",
        )

        self.recovery_frac_mass_H2O.fix(1)
        self._fixed_perf_vars.append(self.initial_chlorine_demand)
        self._fixed_perf_vars.append(self.contact_time)
        self._fixed_perf_vars.append(self.concentration_time)
        self._fixed_perf_vars.append(self.chlorine_decay_rate)

        self.chlorine_dose = Var(
            self.flowsheet().time, units=pyunits.mg / pyunits.L, doc="Chlorine dose"
        )

        @self.Constraint(self.flowsheet().time, doc="Chlorine dose constraint")
        def chlorine_dose_constraint(b, t):
            return b.chlorine_dose[t] == self.initial_chlorine_demand[
                t
            ] + self.chlorine_decay_rate[t] * self.contact_time[t] + (
                self.concentration_time[t]
                / pyunits.convert(self.contact_time[t], to_units=pyunits.minute)
            )

        self._perf_var_dict["Chlorine Dose (mg/L)"] = self.chlorine_dose
        self._perf_var_dict[
            "Initial Chlorine Demand (mg/L)"
        ] = self.initial_chlorine_demand
        self._perf_var_dict["Contact Time (hr)"] = self.contact_time
        self._perf_var_dict["CT Value ((mg*min)/L)"] = self.concentration_time
        self._perf_var_dict[
            "Chlorine Decay Rate (mg/(L*hr))"
        ] = self.chlorine_decay_rate

    @property
    def default_costing_method(self):
        return self.cost_chlorination

    @staticmethod
    def cost_chlorination(blk):
        """
        General method for costing chlorination units. Capital cost is based on
        the both inlet flow and dosage of chlorine.
        This method also registers the chemical flow and electricity demand as
        costed flows.
        """
        t0 = blk.flowsheet().time.first()
        chem_flow_mass = (
            blk.unit_model.chlorine_dose[t0] * blk.unit_model.properties_in[t0].flow_vol
        )

        # Get parameter dict from database
        parameter_dict = blk.unit_model.config.database.get_unit_operation_parameters(
            blk.unit_model._tech_type, subtype=blk.unit_model.config.process_subtype
        )

        # Get costing parameter sub-block for this technology
        A, B, C = blk.unit_model._get_tech_parameters(
            blk,
            parameter_dict,
            blk.unit_model.config.process_subtype,
            ["capital_a_parameter", "capital_b_parameter", "capital_c_parameter"],
        )

        # Determine if a costing factor is required
        factor = parameter_dict["capital_cost"]["cost_factor"]

        # Add cost variable and constraint
        blk.capital_cost = pyo.Var(
            initialize=1,
            units=blk.config.flowsheet_costing_block.base_currency,
            bounds=(0, None),
            doc="Capital cost of unit operation",
        )

        ln_Q = pyo.log(
            pyo.units.convert(
                blk.unit_model.properties_in[t0].flow_vol
                / (pyo.units.m**3 / pyo.units.hour),
                to_units=pyo.units.dimensionless,
            )
        )
        ln_D = pyo.log(
            pyo.units.convert(
                blk.unit_model.chlorine_dose[t0] / (pyo.units.mg / pyo.units.liter),
                to_units=pyo.units.dimensionless,
            )
        )

        expr = pyo.units.convert(
            A * ln_Q + B * ln_D + C * ln_Q * ln_D,
            to_units=blk.config.flowsheet_costing_block.base_currency,
        )

        blk.costing_package.add_cost_factor(
            blk, parameter_dict["capital_cost"]["cost_factor"]
        )

        blk.capital_cost_constraint = pyo.Constraint(
            expr=blk.capital_cost == blk.cost_factor * expr
        )

        # Register flows
        blk.config.flowsheet_costing_block.cost_flow(
            blk.unit_model.electricity[t0], "electricity"
        )
        blk.config.flowsheet_costing_block.cost_flow(chem_flow_mass, "chlorine")
